####################################################################################
# MIT License
#
# Copyright (c) 2023 bluewhitep
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
####################################################################################
from pprint import pprint

from notion_kit import object
from notion_kit.CONTENTS import (
                        NON_CREATEABLE_PROPERTIES_TYPES,
                        NON_UPDATABLE_PROPERTIES_TYPES,
                        NON_UPDATABLE_PROPERTIES_ITEMS,
                        )

class Base_api:
    def __init__(self, client, id:str | None):
        self.id = id
        self.client = client
    
class Page(Base_api):
    def __init__(self, client, id:str | None):
        super().__init__(client, id)
        # if id is not None:
        #     self.page_dict = self.get_data(id)
        #     self.page_info = self.gadget.get_info(self.page_dict)
        #     self.title = self.page_info['name']
        #     self.id = id
        #     self.properties = self.page_info['properties']
    
    def create_in_database(self, parent_database_id:str,
                                 title:object.RichText | None ,
                                 properties_item_dict:dict[str,object.PropertyItem] | None = None,
                                 icon:object.Icon | None = None,
                                 cover:object.FileLink | None = None
                                 # children:list[object.Block] | None = None
                                 ) ->object.Page:
        """
        Create page in database
        
        Parameters:
            parent_database_id:        (str)                    - Target database id
            title:                     (object.RichText)        - New page title
            icon:                      (object.Icon)            - New page icon [Optional]
            cover:                     (object.FileLink)        - New page cover [Optional]
            children:                  (list[object.Block])     - New page children [unavailable]
            
        Returns:
            object.Page:               (object.Page)            - New page
        """
        kwargs = {
            'parent': object.Parent(type="database_id", database_id=parent_database_id).Dict,  # type: ignore
            'properties': {"Name": {"title": [title.Dict]}}                      # type: ignore
            }
        
        if properties_item_dict is not None:
            for name, properties in properties_item_dict.items():
                if properties.type in NON_UPDATABLE_PROPERTIES_ITEMS:
                    continue
                else:
                    kwargs['properties'].update({name: properties.Dict})
                    
        if icon is not None:
            kwargs['icon'] = icon.Dict
        if cover is not None:
            kwargs['cover'] = cover.Dict
        return object.Page(**self.client.pages.create(**kwargs))
   
    def create_in_page(self, parent_page_id, title:object.RichText | None = None,
                                            icon:object.Icon | None = None,
                                            cover:object.FileLink | None = None,
                                            # children:list[object.Block] | None = None
                                            ) ->object.Page:
        """
        Create page in database
        
        Parameters:
            parent_page_id:            (str)                    - Target database id
            title:                     (object.RichText)        - New page title
            icon:                      (object.Icon)            - New page icon [Optional]
            cover:                     (object.FileLink)        - New page cover [Optional]
            children:                  (list[object.Block])     - New page children [unavailable]
            
        Returns:
            object.Page:               (object.Page)            - New page
        """
        kwargs = {
            'parent': object.Parent(type="page_id", page_id=parent_page_id).Dict, # type: ignore
            'properties': {"title": [title.Dict]}, # type: ignore
            }
        if icon is not None:
            kwargs['icon'] = icon.Dict
        if cover is not None:
            kwargs['cover'] = cover.Dict
        return object.Page(**self.client.pages.create(**kwargs))
     
    def get_data(self, page_id:str) ->object.Page:
        """
        Use notion client api get page info dict

        Parameters:
            page_id:            (str)            - Target page id
            
        Return:
            object.Page:        (object.Page)    - Target page information
        """   
        return object.Page(**self.client.pages.retrieve(page_id))
    
    def update(self, new_page_object:object.Page) ->object.Page:
        """
        Update page 

        Parameters:
            new_page_object:        (object.Page)       - Will update dict of properties
            
        Return:
            object.Page:            (object.Page)       - New page dict
        """
        page_dict = new_page_object.Dict
        for name, value in page_dict['properties'].items():
            if value is None:
                continue
            if value['type'] in NON_UPDATABLE_PROPERTIES_ITEMS:
                new_page_object.properties.pop(name)
        new_page_object.update()
        
        return object.Page(**self.client.pages.update(new_page_object.id, 
                                                      **new_page_object.Dict))
    
class Database(Base_api):
    def __init__(self, client, id:str | None):
        super().__init__(client, id)
        # if id is not None:
        #     self.database_dict = self.get_data(id)
        #     self.page_dict = self.get_pages(id)
        #     self.properties_list = Database_gadget.get_properties_options_dict(self.database_dict)
        #     self.pages_list = Database_gadget.get_pages_list(self.page_dict)
        #     self.id = id
        #     self.title = self.database_dict['title'][0]['plain_text']

    def create(self, parent_page_id:str, title:object.RichText,
                                         properties_type_list:list[object.PropertyType] | None = None,
                                         icon:object.Icon | None = None,
                                         cover:object.FileLink | None = None,
                                         is_inline:bool = False) ->object.Database:
        """
        Create database
    
        Parameters:
            parent_page_id:            (str)                       - Target database id
            title:                     (object.RichText)           - New database title
            properties_type:           (object.PropertyType)       - New database properties type [Optional]
            icon:                      (object.Icon)               - New database icon [Optional]
            cover:                     (object.FileLink)           - New database cover [Optional]
            is_inline:                 (bool)                      - New database inline [Optional]
                
        Return:
            object.Database:           (object.Database)           - New database
        """
        
        kwargs = {
            'parent': object.Parent(page_id=parent_page_id, type='page_id').asdict(),
            'title': [title.Dict],
            'properties': {"Name": {"title": {}}}
        }

        if properties_type_list is not None:
            for properties in properties_type_list:
                if properties.type in NON_CREATEABLE_PROPERTIES_TYPES:
                    continue
                else:
                    kwargs['properties'].update(properties.full_dict())
        if icon is not None:
            kwargs['icon'] = icon.Dict
        if cover is not None:
            kwargs['cover'] = cover.Dict
        if is_inline:
            kwargs['is_inline'] = is_inline
        return object.Database(**self.client.databases.create(**kwargs))
    
    def get_pages(self, database_id:str) ->object.DatabaseContainer:
        """
        Get database information

        Parameters:
            database_id:                (str)                       - Target database id
        
        Returns:
            object.DatabaseContainer:  (object.DatabaseContainer)   - Page information of database
        """
        return object.DatabaseContainer(**self.client.databases.query(database_id))

    def get_data(self, database_id:str) ->object.Database:
        """
        Get database information

        Parameters:
            database_id:        (str)       - Target database id

        Returns:
            object.Database:    (object.Database)   - Database information
        """
        return object.Database(**self.client.databases.retrieve(database_id))
        
    def update(self, new_database_object:object.Database) ->object.Database:
        """
        Update database 

        Parameters:
            new_database_object:        (object.Database)       - Will update dict of properties
            
        Return:
            object.Database:            (object.Database)       - New database dict
        """
        database_dict = new_database_object.Dict
        for name, value in database_dict['properties'].items():
            if value is None:
                continue
            if value['type'] in NON_UPDATABLE_PROPERTIES_TYPES:
                new_database_object.properties.pop(name)
        new_database_object.update()
        
        return object.Database(**self.client.databases.update(new_database_object.id, 
                                                              **new_database_object.Dict))
    
class User(Base_api):
    def __dict_to_object(self, user_dict:dict) ->object.User | object.Bot | None:
        """
        Convert dict to object
        
        Parameters:
            user_dict:      (dict)                      - User dict
            
        Returns:
            user_object:    (object.User | object.Bot)  - User object
        """
        if user_dict['type'] == 'person':
            return object.User(**user_dict)
        elif user_dict['type'] == 'bot':
            return object.Bot(**user_dict)
        return None
    
    def get_user_list(self) ->list[object.User | object.Bot]:
        """
        Get user list

        Returns:
            user_list:      (list)      - User list
        """
        user_list = self.client.users.list()['results']
        for i, user in enumerate(user_list):
            user_list[i] = self.__dict_to_object(user)
        return user_list
    
    def get_user_data(self, user_id:str) ->object.User | object.Bot | None:
        """
        Get user data

        Parameters:
            user_id:        (str)                               - Target user id

        Returns:
            user_data:      (object.User | object.Bot | None)   - User data [None if user not found]
        """
        print(self.client.users.retrieve(user_id=user_id))
        return  self.__dict_to_object(self.client.users.retrieve(user_id=user_id))
    
    def who_am_i(self) ->object.Bot:
        """
        Get api user
        Who am I?

        Returns:
            object.Bot:     (object.Bot)    - Api user
        """
        return object.Bot(**self.client.users.me())

class Block(Base_api):
    def get_data(self, id:str) ->object.Block:
        """
        Get block data

        Parameters:
            block_id:           (str)                   - Target block id or page id
            
        Returns:
            object.BlockBase:   (object.BlockBase)      - Block data
        """
        return object.Block(**self.client.blocks.retrieve(block_id=id))
    
    def get_children_blocks(self, id:str) ->object.BlockList:
        """
        Get block childrens

        Parameters:
            block_id:           (str)                   - Target block id or page id
        
        Returns:
            object.BlockList:   (object.BlockList)      - Block childrens
        """
        return object.BlockList(**self.client.blocks.children.list(block_id=id))
    
    def add_block(self, id:str, block_list:list) ->object.BlockList:
        """
        Append block childrens

        Parameters:
            page_id:       (str)       - Target page id
            block_list:     (list)      - Will append block list

        Returns:
            childrens:      (dict)      - Childrens list
        """
        block_list = [block if type(block) == dict else block.block_base_dict()
                      for block in block_list]
        return object.BlockList(**self.client.blocks.children.append(block_id=id,
                                                                    children=block_list))
    
    def update(self, new_block_object: object.Block) ->object.Block:
        """
        Update block childrens

        Parameters:
            page_id:          (str)           - Target page id
            block_list:       (dict)          - Will update block list

        Returns:
            new_block_dict:      (dict)       - New block dict
        """
        return object.Block(**self.client.blocks.update(block_id=new_block_object.id,
                                                        **new_block_object.block_item()))
    
    def del_block(self, block_id:str) ->object.Block:
        """
        Delete block childrens

        Parameters:
            block_id:       (str)       - Target block id

        Returns:
            deleted_block: (dict)      - New block dict
        """
        return object.Block(**self.client.blocks.delete(block_id=block_id))



####################################################################################
# MIT License
#
# Copyright (c) 2023 bluewhitep
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
####################################################################################
import sys
import time, datetime, pytz

from typing import Any
from notion_client.helpers import get_id, is_full_page, is_full_database

from .object import *
from .CONTENTS import (
    TEXT_COLOR_LIST, 
    ROLLUP_FUNCTION_LIST, 
    BLOCK_TYPE_LIST,
    CODE_LANGUAGE_LIST,
    MEDIA_TYPE_LIST,
    NON_CREATEABLE_PROPERTIES_TYPES,
    )

class Property_gadget:
    @staticmethod
    def is_exist(prop_dict:dict, target:list) ->bool:
        """
        Check if the prop_dict is exist in dict of list
        
        Parameters:
            prop_dict:          (dict)      - Target dict to check
            target:             (list)      - Target list to check

        Returns:
            result:             (bool)      - Result
        """
        for target_dict in target:
            # Check 'name' and 'id' key
            if (('name' in prop_dict) and (prop_dict['name'] == target_dict['name'])) or \
                (('id' in prop_dict) and (prop_dict['id'] == target_dict['id'])):
                 return True
        return False

    @staticmethod
    def get_properties_name(database_properties_dict:dict, type:str) ->list:
        """
        Get properties name

        Parameters:
            database_properties_dict:    (dict)      - Target properties dict of database
            type:                        (str)       - Target properties type

        Returns:
            (list)  - Target property names
        """
        keys = []
        for key, item in database_properties_dict.items():
            if item['type'] == type:
                keys.append(key)
        return keys
    
    @staticmethod
    def get_property_dict(database_properties_dict:dict, name:str) ->dict:
        """
        Get property dict

        Parameters:
            database_properties_dict:    (dict)      - Target properties dict of database
            name:                        (str)       - Target properties name

        Returns:
            (dict)  - Target property dict
        """
        return {name: database_properties_dict[name]}
    
    @staticmethod
    def del_properties(database_properties_dict:dict, names:list) -> None:
        """
        Delete properties

        Parameters:
            database_properties_dict:    (dict)      - Target properties dict of database
            name:                        (str)       - Target properties name

        Returns:
            (dict)  - Target properties dict
        """
        if not (names == []):
            for name in names:
                database_properties_dict.pop(name)
    
    # FC: [Get property type dict]
    @staticmethod
    def create_dict(name:str, type:str, **kwargs) ->dict:
        """
        Create properties

        Parameters:
            name:           (str)       - Properties name
            type:           (str)       - Properties type
            **kwargs:       (dict)      - Properties options
                exsampe:
                    format:     (str)       - Number format
                    options:    (list)      - Select and Multi-select options
                    expression: (str)       - Formula expression
                       
        Return:
            properties:     (dict)      - Properties dict
            
        Raises:
            Exception:      (Exception) - Error type value
        """
        if type == 'rich_text':
            # return {name: {'rich_text': {},
            #                'name': name,
            #                'type': 'rich_text'}}
            return PropertyType(name=name, type=type).full_dict()
        elif type == 'number':
            # if kwargs == {}:
            #     kwargs = {'format': 'number'}
            # return {name: {'number': kwargs,
            #                'name': name,
            #                'type': 'number'}}
            return PropertyType(name=name, type=type,
                                number=Number() if kwargs == {} else Number(**kwargs)
                                ).full_dict()            
        elif type == 'select':
            # if kwargs == {}:
            #     kwargs = {'options': []}
            # return {name: {'select': kwargs,
            #                'name': name,
            #                'type': 'select'}}
            return PropertyType(name=name, type=type,
                                select=Options() if kwargs == {} else Options(**kwargs)
                                ).full_dict()
        elif type == 'multi_select':
            # if kwargs == {}:
            #     kwargs = {'options': []}
            # return {name: {'multi_select': kwargs,
            #                 'name': name,
            #                 'type': 'multi_select'}}
            return PropertyType(name=name, type=type,
                                multi_select=Options() if kwargs == {} else Options(**kwargs)
                                ).full_dict()
        elif type == 'date':
            # return {name: {'date': {},
            #                'name': name,
            #                'type': 'date'}}
            return PropertyType(name=name, type=type).full_dict()
        elif type == 'people':
            # return {name: {'people': {},
            #                'name': name,
            #                'type': 'people'}}
            return PropertyType(name=name, type=type).full_dict()
        elif type == 'files':
            # return {name: {'files': {},
            #                 'name': name,
            #                 'type': 'files'}}
            return PropertyType(name=name, type=type).full_dict()
        elif type == 'checkbox':
            # return {name: {'checkbox': {},
            #                'name': name,
            #                'type': 'checkbox'}}
            return PropertyType(name=name, type=type).full_dict()
        elif type == 'url':
            # return {name: {'url': {},
            #                'name': name,
            #                'type': 'url'}}
            return PropertyType(name=name, type=type).full_dict()
        elif type == 'email':
            # return {name: {'email': {},
            #                'name': name,
            #                'type': 'email'}}
            return PropertyType(name=name, type=type).full_dict()
        elif type == 'phone_number':
            # return {name: {'phone_number': {},
            #                'name': name,
            #                'type': 'phone_number'}}
            return PropertyType(name=name, type=type).full_dict()
        elif type == 'formula':
            # if kwargs == {}:
            #     kwargs = {'expression': ''}
            # return {name: {'formula': kwargs,
            #                'name': name,
            #                'type': 'formula'}}
            return PropertyType(name=name, type=type,
                                formula=Formula() if kwargs == {} else Formula(**kwargs)).full_dict()
        elif type == 'relation':
            if kwargs == {} or not ('relation' in kwargs):
                raise Exception(f"{sys._getframe().f_code.co_name}: " +
                                 "Parameter error, please check. " +
                                 "Relation properties need to set 'relation' key." +
                                 "Recommend to use 'gadget.create_relation_dict' method.")
            # return {name: {'relation': kwargs['relation_dict'],
            #                'name': name,
            #                'type': 'relation'}}
            return PropertyType(name=name, type=type,
                                relation=kwargs['relation']).full_dict()
        elif type == 'rollup':
            if kwargs == {} or not ('rollup_dict' in kwargs):
                raise Exception(f"{sys._getframe().f_code.co_name}: " +
                                 "Parameter error, please check. " +
                                 "Rollup properties need to set 'rollup_dict' key." +
                                 "Recommend to use 'gadget.create_rollup_dict' method.")
            # return {name: {'rollup': kwargs['rollup_dict'],
            #                'name': name,
            #                'type': 'rollup'}}
            return PropertyType(name=name, type=type,
                                rollup=kwargs['rollup']).full_dict()
        elif type == 'created_time':
            # return {name: {'created_time': {},
            #                'name': name,
            #                'type': 'created_time'}}
            return PropertyType(name=name, type=type).full_dict()
        elif type == 'created_by':
            # return {name: {'created_by': {},
            #                'name': name,
            #                'type': 'created_by'}}
            return PropertyType(name=name, type=type).full_dict()
        elif type == 'last_edited_time':
            # return {name: {'last_edited_time': {},
            #                'name': name,
            #                'type': 'last_edited_time'}}
            return PropertyType(name=name, type=type).full_dict()
        elif type == 'last_edited_by':
            # return {name: {'last_edited_by': {},
            #                'name': name,
            #                'type': 'last_edited_by'}}
            return PropertyType(name=name, type=type).full_dict()
        elif type == 'status':
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter. " +
                            "'status' is not support.")
        else:
            raise Exception(f"{sys._getframe().f_code.co_name}: " + 
                            "Parameter error, please check the parameter. " +
                            "'type' is not in 'rich_text', 'number', 'select', 'multi_select', 'date', 'people', 'file', 'checkbox', 'url', 'email', 'phone_number', 'formula', 'relation', 'rollup', 'created_time', 'created_by', 'last_edited_time', 'last_edited_by'")
    
    ## Property
    ### Set property name
    @staticmethod
    def set_property_name(property_dict:dict, name:str) ->dict:
        """
        Set property name

        Parameters:
            property_dict:        (dict)      - property dict
            name:                 (str)       - New name
            
        Return:
            new_property_dict:    (dict)      - New property dict
        """
        for _,items in property_dict.items():
            items['name'] = name
        return property_dict
    
    ### Rename property name
    @staticmethod
    def rename_property(old_name:str, new_name:str) ->dict:
        """
        Rename property name

        Parameters:
            old_name:        (str)       - Old name
            new_name:        (str)       - New name
            
        Return:
            property_dict:   (dict)      - New property dict
        """
        return {old_name:{'name':new_name}}
        
    ### Mix properties
    @staticmethod
    def mix_properties(*args) ->dict:
        """
        Mix properties

        Parameters:
            *args:        (dict)      - properties dict
            
        Return:
            properties:   (dict)      - New properties dict
        """
        properties = {}
        for arg in args:
            properties.update(arg)
        return properties
      
    ### Text
    #### Page
    @staticmethod
    def set_text(name:str, text:str, **kwargs) ->dict:
        """
        Create new text properties dict

        Args:
            name:     (str)              - Name of text properties
            text:     (list or dict)     - properties
            **kwargs: (dict)             - Text style

        Returns:
            text:     (dict)             - new text properties dict
        """
        new_text =  {'text': {'content': text}}
        if 'style' in kwargs:
            new_text.update({'annotations':kwargs['style']})
        else:
            new_text.update({'annotations':Property_gadget.set_text_style(**kwargs)})
        return {name: {'rich_text': [new_text]}}

    @staticmethod
    def set_text_style(bold:bool=False, code:bool=False,
                       color:str='default', italic:bool=False,
                       strikethrough:bool=False, underline:bool=False,)->dict:
        return {'bold': bold, 'code': code, 'color': color, 'italic': italic, 'strikethrough': strikethrough, 'underline': underline}
    
    @staticmethod
    def update_text_style(text:dict, style:dict) ->dict:
        """
        Mix text style

        Parameters:
            text:          (dict)      - Text dict
            style:         (dict)      - Style dict
            
        Return:
            text:          (dict)      - Text dict
        """
        for _, item in text.items():
            item['rich_text'][0]['annotations'] = style
        return text
    
    @staticmethod
    def add_next_line(text:dict) ->dict:
        """
        Set next line
        
        Parameters:
            text:          (dict)      - Text dict
            
        Return:
            text:          (dict)      - Text dict
        """
        next_line = {'annotations': {'bold': False,
                                'code': False,
                                'color': 'default',
                                'italic': False,
                                'strikethrough': False,
                                'underline': False},
                'href': None,
                'plain_text': '\n',
                'text': {'content': '\n', 'link': None},
                'type': 'text'}
        for _, item in text.items():
            item['rich_text'].append(next_line)
        return text
    
    @staticmethod
    def add_new_text(text_dict:dict, new_text:str, **kwargs) ->dict:
        """
        Add new text

        Parameters:
            text_dict:     (dict)      - Text dict
            new_text:      (str)       - New text
            **kwargs:      (dict)      - Text style
            
        Return:
            text_dict:     (dict)      - Text dict
        """
        new_text_dict =  {'text': {'content': new_text}}
        new_text_dict.update({'annotations':Property_gadget.set_text_style(**kwargs)})
        
        for _, item in text_dict.items():
            item['rich_text'].append(new_text_dict)
        return text_dict
    
    @staticmethod
    def add_metion(text_dict:dict, dict:dict, **kwargs) ->dict:
        """
        Add metion

        Parameters:
            text_dict:     (dict)      - Text dict
            dict:          (dict)      - Metion dict is maybe page or database
            **kwargs:      (dict)      - Text style
            
        Return:
            text_dict:     (dict)      - Text dict
            
        Raises:
            Exception:     (Exception) - dict is not page or database
        """
        type = Dict_Gadget.check_dict_type(dict)
        if type == 'none':
            raise Exception('dict is not page or database')
        
        metion_dict= {'href': 'https://www.notion.so/'+dict['id'],
                      'type': 'mention'}
        if type == 'page':
            metion_dict.update({'mention': {'page': {'id': dict['id']},
                                            'type': 'page'},
                                'plain_text': Page_gadget.get_title(dict)})
        elif type == 'database':
            metion_dict.update({'mention': {'database': {'id': dict['id']},
                                            'type': 'database'},
                                'plain_text': Database_gadget.get_title(dict)})
        
        metion_dict.update({'annotations':Property_gadget.set_text_style(**kwargs)})
        for _, item in text_dict.items():
            item['rich_text'].append(metion_dict)
        return text_dict
    
    @staticmethod
    def add_hyperlink(text_dict:dict, text:str, link:str, **kwargs) ->dict:
        """
        Add hyperlink

        Parameters:
            text_dict:     (dict)      - Text dict
            text:          (str)       - Text
            link:          (str)       - Hyperlink url
            **kwargs:      (dict)      - Text style
            
        Return:
            text_dict:     (dict)      - Text dict
            
        Raises:
            Exception:     (Exception) - dict is not page or database
        """
        hyperlink_dict = {
                          'text':{'content': text,
                                  'link': {'url': link}},
                          'type': 'text'}
        hyperlink_dict.update({'annotations':Property_gadget.set_text_style(**kwargs)})
        for _, item in text_dict.items():
            item['rich_text'].append(hyperlink_dict)
        return text_dict
        
    @staticmethod
    def mix_text(*args) ->dict:
        """
        Mix text

        Parameters:
            *args:        (dict)      - Text dict   
            
        Return:
            text:          (dict)      - Text dict
        """
        text = args[0]
        new_text = args[1]
        
        new_text_dict = {}
        for _, item in new_text.items():
            new_text_dict = item['rich_text'][0]
        for _, item in text.items():
            item['rich_text'].append(new_text_dict)
        return text
    
    ### Checkbox
    @staticmethod
    def set_checkbox(name:str, checked:bool) ->dict:
        """
        Set checkbox

        Parameters:
            name:               (str)       - Name of checkbox property
            checked:            (bool)      - Checkbox checked
            
        Return:
            checkbox_dict:      (dict)      - Checkbox dict
        """
        return {name: {'checkbox': checked}}
    
    ### Number
    #### Database
    @staticmethod
    def set_number_format(number_dict:dict, format:str) ->dict:
        """
        Set number format

        Parameters:
            number_dict:        (dict)      - Number dict
            format:             (str)       - Number format
            
        Return:
            new_number_dict:    (dict)      - New number dict
        """
        for _, item in number_dict.items():
            item['number']['format'] = format
        return number_dict
    
    #### Page
    @staticmethod
    def set_number(name:str, number:float) ->dict:
        """
        Set number

        Parameters:
            name:               (str)       - Name of number property
            number:             (float)     - Number
            
        Return:
            number_dict:        (dict)      - Number dict
        """
        return {name: {'number': number}}
    
    ### Select
    #### Function
    @staticmethod
    def check_select_option(options_list:list, option:str) ->bool:
        """
        Check select option if in options list

        Args:
            options_list    (list):     - Select options list
            option          (str):      - Select option

        Returns:
            bool:                       - True if in options list, else False
        """
        if option in options_list:
            return True
        else:
            return False
    
    @staticmethod
    def set_select_option(option_name:list, option_color:list) ->list:
        """
        Set select option

        Parameters:
            option_name:            (list)      - Option name list
            option_color:           (list)      - Option color list
            
        Return:
            option_list:            (list)      - Option list
        """
        option_list = []
        for i in range(len(option_name)):
            option_list.append({'name': option_name[i],
                                'color': option_color[i]})
        return option_list
    
    #### Database
    @staticmethod
    def get_select_options_list(database_property_dict:dict, select_name:str) ->list:
        """
        Set select options list

        Parameters:
            database_property_dict:        (dict)      - property dict
            select_name:                   (str)       - select name
            
        Return:
            options_list:                  (list)      - select options list
        """
        options_list = []
        for option in database_property_dict[select_name]['select']['options']:
            options_list.append(option['name'])
        return options_list
    
    #### Page
    @staticmethod
    def set_select(name:str, option:str) ->dict:
        """
        Set select

        Parameters:
            name:               (str)       - Name of select property
            option:             (str)       - Select dict
            
        Return:
            select_dict:        (dict)      - Select dict
        """
        return {name: {'select':  {'name': option}}}
    
    ### Multi-Select
    #### Database
    @staticmethod
    def get_multi_select_options_list(database_property_dict:dict, select_name:str) ->list:
        """
        Set multi_select options list

        Parameters:
            database_property_dict:        (dict)      - property dict
            select_name:                   (str)       - select name
            
        Return:
            options_list:                  (list)      - select options list
        """
        options_list = []
        for option in database_property_dict[select_name]['multi_select']['options']:
            options_list.append(option['name'])
        return options_list
    
    #### Page
    @staticmethod
    def set_multi_select(name:str, options:list) ->dict:
        """
        Set multi_select

        Parameters:
            name:               (str)       - Name of select property
            options:            (list)      - Select dict
            
        Return:
            multi_select_dict:  (dict)      - Select dict
        """
        multi_select_dict =  {name: {'multi_select': []}}
        for option in options:
            multi_select_dict[name]['multi_select'].append({'name': option})
        return multi_select_dict
    
    ### Date
    @staticmethod
    def set_date(name:str, start_date:str, end_date:Any=None) ->dict:
        """
        Set date

        Parameters:
            name:               (str)       - Name of date property ex: 'YYYY-MM-DDThh:mm:ss.000+09:00'
            start_date:         (str)       - Start date ex: 'YYYY-MM-DDThh:mm:ss.000+09:00'
            end_date:           (str)       - End date
            
        Return:
            date_dict:          (dict)      - Date dict
        """
        return {name: {'date': {'start': start_date, 'end': end_date, 'time_zone': None}}}
    
    ### Status
    @staticmethod
    def get_status_options_dict(database_property_dict:dict, status_name:str) ->dict:
        """
        Get status options dict

        Parameters:
            database_property_dict:        (dict)      - property dict
            status_name:                   (str)       - status name
            
        Return:
            options_dict:                  (dict)      - status options dict
                Ex:
                    {'group name1': [{'color': 'gray', 'name': 'Not Started'},
                                {'color': 'yellow', 'name': 'Todo'}],
                    'group name2': [{'color': 'green', 'name': 'Done'}]}
            
        """
        options_dict = {}
        # Get group name
        for group in database_property_dict[status_name]['status']['groups']:
            options_dict.update({group['name']: group['option_ids']})
        
        # Get options
        for option in database_property_dict[status_name]['status']['options']:
            for _,ids in options_dict.items():
                if option['id'] in ids:
                    ids.remove(option['id'])
                    ids.append(option['name'])
            
        return options_dict
    
    @staticmethod
    def set_status(name:str, option:str) ->dict:
        """
        Set status

        Parameters:
            name:               (str)       - Name of status property
            option:             (str)       - Status option
            
        Return:
            status_dict:        (dict)      - Status dict
        """
        return {name: {'status': {'name': option}}}
    
    ### Email
    @staticmethod
    def set_email(name:str, email:str) ->dict:
        """
        Set email

        Parameters:
            name:               (str)       - Name of email property
            email:              (str)       - Email
            
        Return:
            email_dict:         (dict)      - Email dict
        """
        return {name: {'email': email}}
    
    ### Files
    @staticmethod
    def get_files_path(page_property_dict:dict, files_name:str) ->list:
        """
        Get files path

        Parameters:
            page_property_dict:        (dict)      - property dict
            files_name:                (str)       - files property name
            
        Return:
            files_path:                (list)      - files path
        """
        files_path = []
        for file in page_property_dict[files_name]['files']:
            files_path.append(file)
        return files_path
    
    @staticmethod
    def del_files(name:str) ->dict:
        """
        Delete files

        Parameters:
            name:               (str)       - Name of files property
            
        Return:
            files_dict:         (dict)      - Files dict
        """
        return {name: {'files': []}}
    
    @staticmethod
    def pass_files(name:str, files_path:list) ->dict:
        """
        Pass files

        Parameters:
            name:               (str)       - Name of files property
            files_path:         (list)      - Files path
            
        Return:
            files_dict:         (dict)      - Files dict
        """
        # check path
        for path in files_path:
            # check file
            if 'type' in path \
                and ((path['type'] == 'file' and 'file' in path \
                        and 'expiry_time' in path['file'] \
                        and 'url' in path['file']) \
                    or (path['type'] == 'external' and 'external'in path \
                        and 'url' in path['external'])):
                raise Exception(f"{sys._getframe().f_code.co_name}: " +
                                "Parameter error, please check the parameter. " +
                                "'files_path' is not form notion api. " +
                                "(https://developers.notion.com/reference/file-object)")
                
        return {name: {'files': files_path}}
    
    ### Phone Number
    @staticmethod
    def set_phone_number(name:str, phone_number:str) ->dict:
        """
        Set phone number

        Parameters:
            name:               (str)       - Name of phone number property
            phone_number:       (str)       - Phone number
            
        Return:
            phone_number_dict:  (dict)      - Phone number dict
        """
        return {name: {'phone_number': phone_number}}
    
    ### People
    @staticmethod
    def set_people(name:str, people_dicts:list) ->dict:
        """
        Set people

        Parameters:
            name:               (str)       - Name of people property
            people_dicts:       (list)      - People dict of list
            
        Return:
            people_dict:        (dict)      - People dict
        """
        return {name: {'people': people_dicts}}
    
    @staticmethod
    def del_pepole(name:str) ->dict:
        """
        Delete people

        Parameters:
            name:               (str)       - Name of people property
            
        Return:
            people_dict:        (dict)      - People dict
        """
        return {name: {'people': []}}
    
    ### Url
    @staticmethod
    def set_url(name:str, url:str) ->dict:
        """
        Set url

        Parameters:
            name:               (str)       - Name of url property
            url:                (str)       - Url
            
        Return:
            url_dict:           (dict)      - Url dict
        """
        return {name: {'url': url}}
    
    ### Rollup
    @staticmethod
    def create_rollup_dict(function:str, relation_property_name:str, rollup_property_name:str) ->dict:
        """
        Get rollup dict

        Parameters:
            function:                   (str)       - Rollup function
            relation_property_name:     (str)       - Relation property name
            rollup_property_name:       (str)       - Rollup property name
            
        Return:
            rollup_dict:                (dict)      - Rollup dict
            
        Raise:
            Exception:                  (Exception) - Parameter error
        """
        if not Dict_Gadget.check_rollup_function(function):
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                             "Parameter error, please check the parameter. " +
                            f"'function' is not in ROLLUP_FUNCTION_LIST. \n" +
                            f" ROLLUP_FUNCTION_LIST: {ROLLUP_FUNCTION_LIST}")
        return {'function': function,
                'relation_property_name': relation_property_name,
                'rollup_property_name': rollup_property_name}
        
    @staticmethod
    def get_rollup_value(page_property_dict:dict, rollup_property_name:str) ->dict:
        """
        Get rollup value

        Parameters:
            page_property_dict:         (dict)      - Property dict
            rollup_property_name:       (str)       - Rollup property name
            
        Return:
            rollup_value:               (dict)      - Rollup value
        """
        return page_property_dict[rollup_property_name]['rollup']
    
    ### Relation
    @staticmethod
    def create_relation_dict(database_id:str, is_dual:bool) ->dict:
        """
        Get relation dict

        Parameters:
            database_id:        (str)       - Database id
            is_dual:            (bool)      - If is 'dual_property' will be true
            
        Return:
            relation_dict:      (dict)      - Relation dict
        """
        if is_dual:
            # relation_dict = {'database_id': database_id,
            #                   'type': 'dual_property',
            #                   'dual_property': {},
            #                 }
            relation_dict = RelationDual(database_id=database_id).asdict()
        else:
            # relation_dict = {'database_id': database_id,
            #                  'type': 'single_property',
            #                  'single_property': {},
            #                 }
            relation_dict = RelationSingle(database_id=database_id).asdict()
        return relation_dict
    
    @staticmethod
    def set_relation(name:str, relation_ids:list) ->dict:
        """
        Set relation

        Parameters:
            name:               (str)       - Name of relation property
            relation_id:        (list)       - Relation ids
            
        Return:
            relation_dict:      (dict)      - Relation dict
        """
        ids = []
        for id in relation_ids:
            ids.append({'id': id})
        return {name: {'relation': ids}}
    
class Page_gadget:
    @staticmethod
    def get_id(page_dict:dict) ->str:
        """
        Get page id from raw dict

        Patameters:
            page_dict:          (dict)      - Raw page dict form Client.pages.retrieve()
        
        Return
            page_id:            (str)       - Page id
        """
        if 'object' in page_dict and page_dict['object'] == 'page':
            return page_dict['id']
        else:
            raise Exception(f"{sys._getframe().f_code.co_name}: " + 
                            "Parameter error, please check the parameter. " +
                            "'page_dict' is not form Client.pages.retrieve()")
    
    @staticmethod
    def get_title(page_dict:dict) ->Any:
        """
        Get page title
        
        Parameters:
            page_dict:          (dict)      - Raw page dict form Client.pages.retrieve()   ** Choose one of the two parameters **
        
        Returns:
            title:              (str)       - Target page title
        """
        if 'object' in page_dict and page_dict['object'] == 'page':
            properties_data = page_dict['properties']
        else: 
            raise Exception(f"{sys._getframe().f_code.co_name}:  " +
                            "Parameter error, please check the parameter. " +
                            "'page_dict' is not form Client.pages.retrieve()")

        for item in properties_data.items():
            if item[1]['type'] == 'title':
                return item[1]['title'][0]['plain_text']
            
    @staticmethod
    def get_info(page_dict:dict={}) ->dict:
        """
        Get page information

        Parameters:
            page_dict:          (dict)      - Raw page dict form Client.pages.retrieve()
            
        Returns:
            page_info:          (dict)      - Target page information
                                            - exsample:[{'name': "",
                                                         'id': "",
                                                         'properties: [{'name': "",
                                                                        'value': {}}, {...}],
                                                        }, {...}]
        """
        if not ('object' in page_dict and page_dict['object'] == 'page'): 
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter. " +
                            "'page_dict' is not form Client.pages.retrieve()")
            
        page_info = {'name': Page_gadget.get_title(page_dict=page_dict),
                    'id': page_dict['id'],
                    'properties': page_dict['properties']}
        return page_info
    
    @staticmethod
    def get_properties(page_dict:dict) ->dict:
        """
        Get page properites from page info dict

        Parameters:
            page_dict:          (dict)      - Raw page dict form Client.pages.retrieve()
        
        Return:
            page_properties:    (dict)      - Target page properties
        """
        if 'properties' in page_dict:
            return page_dict['properties']
        else:
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "page dict not include 'property'")
    
class Database_gadget:
    @staticmethod
    def get_id(database_dict:dict) ->str:
        """
        Get database title
        
        Parameters:
            database_dict:      (dict)      - Raw page dict form Client.pages.retrieve()   ** Choose one of the two parameters **
        
        Returns:
            title:              (str)       - Target page title
        """
        if 'object' in database_dict and database_dict['object'] == 'database':
            return database_dict['id']
        else: 
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter. " +
                            "'database_dict' is not form Client.databases.retrieve()")
    
    @staticmethod
    def get_title(database_dict:dict) ->str:
        """
        Get database title
        
        Parameters:
            database_dict:      (dict)      - Raw page dict form Client.pages.retrieve()   ** Choose one of the two parameters **
        
        Returns:
            title:              (str)       - Target page title
        """
        if 'object' in database_dict and database_dict['object'] == 'database':
            return database_dict['title'][0]['plain_text']
        else: 
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter. " +
                            "'database_dict' is not form Client.databases.retrieve()")

    @staticmethod
    def get_pages_list(database_page_dict:dict={}) ->list:
        """
        Get page list in the database of table

        Parameters:
            database:           (dict)      - Target database dict form Client.databases.query()
            
        Returns:
            page_list:          (list)      - Target page list
                                            - exsample:[{'name': "",
                                                         'id': ""}, {...}]                                        
        """
        if 'object' in  database_page_dict \
                    and database_page_dict['object'] == 'list' :
            pages_list = database_page_dict['results']
        else:
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter. " +
                            "'database_page_dict' is not form Client.databases.query()")
        page_list = []
        for page in pages_list:
            page_list.append({'name': Page_gadget.get_title(page),
                              'id':page['id']})
        return page_list
    
    @staticmethod
    def get_properties_options_dict(database_dict:dict={}) ->dict:
        """
        Get properties list in the database of table

        Parameters:
            database_dict:      (dict)      - Target database dict form Client.databases.retrieve()
        Returns:
            properties_dict:    (dict)      - Target properties dict
                                            - exsample:{'name': {'id': "",
                                                                'type': "",
                                                                'value': {}
                                                                }, 
                                                        ...
                                                        }
        """
        if 'object' in  database_dict \
                    and database_dict['object'] == 'database':
            properties = database_dict['properties']
        else:
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter." +
                            "'database_dict' is not form Client.databases.retrieve()")

        properties_dict = {}
        for key, item in properties.items():
            properties_dict.update({key: {'id': item['id'],
                                          'type': item['type'],
                                           item['type']: item[item['type']]}})
            
        return properties_dict
                    
class Block_gadget:
    @staticmethod
    def get_block_ids_and_type(block_dict:dict) ->list:
        """
        Get block ids
        
        Parameters:
            block_dict:         (dict)      - Target block dict
            
        Returns:
            block_infos:          (list)      - Target block ids
                                            - exsample:[{'type': 'page',
                                                        'ids': ['id1']},
                                                        ...]
            
        Raises:
            Exception:          (Exception) - block dict not include 'results
        """
        if 'results' not in block_dict:
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "block dict not include 'results'")
        blocks_info = []
        for block in block_dict['results']:
            blocks_info.append({'type': block['type'],
                            'id': block['id']})
        return blocks_info
    
    @staticmethod
    def get_block_id(blocks_info:list, type:str) ->str:
        """
        Take block id
        
        Parameters:
            blocks_info:        (list)      - Target block info list
            
        Returns:
            block_id:           (str)       - Target block id
        """
        if type == 'text':
            type = 'paragraph'
        for block_info in blocks_info:
            if block_info['type'] == type:
                return block_info['id']
        return ""
    
    @staticmethod
    def get_generic_block_dict(type:str ,**kwargs) ->dict:
        """
        Get block dict
        
        Parameters:
            type:               (str)       - Target block type
            kwargs:             (dict)      - Target block data
            
        Returns:
            block_dict:         (dict)      - Target block dict
        """
        if not Dict_Gadget.check_block_type(type):
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter. " +
                            "'type' is not in BLOCK_TYPE_LIST. \n" +
                            f"BLOCK_TYPE_LIST = {BLOCK_TYPE_LIST}")
        block_dict = {'object': 'block',
                      'type': type,
                    }
        
        if (type == 'heading_1' or
            type == 'heading_2' or
            type == 'heading_3'):
            block_dict.update(kwargs['text'])
            kwargs.pop('text')
            if 'is_toggleable' in kwargs:
                block_dict[type].update({'is_toggleable': kwargs['is_toggleable']})  # type: ignore "block_dict[type] is dict"
                kwargs.pop('is_toggleable')
        
        block_dict.update(kwargs)
        return block_dict
 
    # FC: [Block dict] paragraph
    @staticmethod
    def get_text_block(text:str, **kwargs) ->dict:
        """
        Get paragraph block dict
        
        Parameters:
            text:               (str)       - Target block text
            kwargs:             (dict)      - Target block data
            
        Returns:
            paragraph_block:    (dict)      - Target block dict
        """
        type = 'paragraph'
        paragraph_text = {'text': {'content': text}}
        if 'style' in kwargs:
            paragraph_text.update({'annotations':kwargs['style']})
        else:
            paragraph_text.update({'annotations':Property_gadget.set_text_style(**kwargs)})
        
        paragraph_block = {'type': type,
                           type:{'rich_text': [paragraph_text]},
                        }
        
        return paragraph_block    
    
    # FC: [Block dict] heading
    @staticmethod
    def get_heading_block(text:str, type:str, is_toggleable:bool=False, **kwargs) ->dict:
        """
        Get heading block dict
        
        Parameters:
            text:               (str)       - Target block text
            type:               (str)       - Target block type
            is_toggleable:      (bool)      - Target block is toggleable
            kwargs:             (dict)      - Target block data
            
        Returns:
            heading_block:         (dict)      - Target block dict
            
        Raises:
            Exception:          (Exception) - Parameter error
        """        
        if not (type == 'heading_1' or
                type == 'heading_2' or
                type == 'heading_3'):
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter. " +
                            "'type' is not in ['heading_1', 'heading_2', 'heading_3'].")
        
        title_text = {'text': {'content': text}}
        if 'style' in kwargs:
            title_text.update({'annotations':kwargs['style']})
        else:
            title_text.update({'annotations':Property_gadget.set_text_style(**kwargs)})
        heading_block = {'type': type,
                        type:{'rich_text': [title_text],
                              'is_toggleable': is_toggleable,
                              },
                        }
        return heading_block
    
    # FC: [Block dict] callout
    @staticmethod
    def get_callout_block(text:str, icon:str="📦", **kwargs) ->dict:
        """
        Get callout block dict
        
        Parameters:
            text:               (str)       - Target block text
            icon:               (str)       - Target block icon
            kwargs:             (dict)      - Target block data
            
        Returns:
            callout_block:         (dict)      - Target block dict
        """
        type = 'callout'
        callout_text = {'text': {'content': text}}
        if 'style' in kwargs:
            callout_text.update({'annotations':kwargs['style']})
        else:
            callout_text.update({'annotations':Property_gadget.set_text_style(**kwargs)})
        
        callout_block = {'type': type,
                         type: {
                                'rich_text': [callout_text]
                                },
                        }
        
        return callout_block  
        
    # FC: [Block dict] quote list
    @staticmethod
    def get_quote_block(text:str, color:str="default", **kwargs) ->dict:
        """
        Get quote block dict
        
        Parameters:
            text:               (str)        - Target block text
            kwargs:             (dict)       - Target block data
            
        Returns:
            quote_block:        (dict)       - Target block dict
            
        Raises:
            Exception:          (Exception)  - Parameter error
        """
        if not Dict_Gadget.check_text_color(color):
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter. " +
                            "'color' is not in TEXT_COLOR_LIST. \n" +
                            f"TEXT_COLOR_LIST = {TEXT_COLOR_LIST}")
            
        type = 'quote'
        quote_text = {'text': {'content': text}}
        if 'style' in kwargs:
            quote_text.update({'annotations':kwargs['style']})
        else:
            quote_text.update({'annotations':Property_gadget.set_text_style(**kwargs)})
        
        quote_block = {'type': type,
                       type: {
                             'rich_text': [quote_text],
                             'color': color,
                             },
                      }
        if "children" in kwargs:
            if type(kwargs["children"]) is list: # type: ignore "kwargs["children"] need be is list"
                quote_block[type].update({'children': kwargs["children"]})
            else:
                raise Exception(f"{sys._getframe().f_code.co_name}: " +
                                "Parameter error, please check the parameter. " +
                                "'children' is not a list.")

        return quote_block
    
    # FC: [Block dict] bulleted list
    @staticmethod
    def get_bulleted_list_block(text:str, color:str="default", **kwargs) ->dict:
        """
        Get bulleted list block dict
        
        Parameters:
            items:                    (list)      - Target block items
            kwargs:                   (dict)      - Target block data
            
        Returns:
            bulleted_list_block:      (dict)      - Target block dict
            
        Raises:
            Exception:                (Exception) - Parameter error
        """
        if not Dict_Gadget.check_text_color(color):
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter. " +
                            "'color' is not in TEXT_COLOR_LIST. \n" +
                            f"TEXT_COLOR_LIST = {TEXT_COLOR_LIST}")
        
        type = 'bulleted_list_item'
        bulleted_list_text = {'text': {'content': text}}
        if 'style' in kwargs:
            bulleted_list_text.update({'annotations':kwargs['style']})
        else:
            bulleted_list_text.update({'annotations':Property_gadget.set_text_style(**kwargs)})
        
        bulleted_list_block = {'type': type,
                                type: {
                                        'rich_text': [bulleted_list_text],
                                        'color': color,
                                      },
                                }

        if "children" in kwargs:
            if type(kwargs["children"]) is list: # type: ignore "kwargs["children"] need be is list"
                bulleted_list_block[type].update({'children': kwargs["children"]})
            else:
                raise Exception(f"{sys._getframe().f_code.co_name}: " +
                                "Parameter error, please check the parameter. " +
                                "'children' is not a list.")

        return bulleted_list_block  
    
    # FC: [Block dict] numbered list
    @staticmethod
    def get_numbered_list_block(text:str, color:str="default", **kwargs) ->dict:
        """
        Get numbered list block dict
        
        Parameters:
            items:                   (list)      - Target block items
            kwargs:                  (dict)      - Target block data
            
        Returns:
            numbered_list_block:     (dict)      - Target block dict
            
        Raises:
            Exception:               (Exception) - Parameter error
        """
        if not Dict_Gadget.check_text_color(color):
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter. " +
                            "'color' is not in TEXT_COLOR_LIST. \n" +
                            f"TEXT_COLOR_LIST = {TEXT_COLOR_LIST}")
        
        type = 'numbered_list_item'
        numbered_list_text = {'text': {'content': text}}
        if 'style' in kwargs:
            numbered_list_text.update({'annotations':kwargs['style']})
        else:
            numbered_list_text.update({'annotations':Property_gadget.set_text_style(**kwargs)})
        
        numbered_list_block = {'type': type,
                                type: {
                                        'rich_text': [numbered_list_text],
                                        'color': color,
                                      },
                                }

        if "children" in kwargs:
            if type(kwargs["children"]) is list:  # type: ignore "kwargs["children"] need be is list"
                numbered_list_block[type].update({'children': kwargs["children"]})
            else:
                raise Exception(f"{sys._getframe().f_code.co_name}: " +
                                "Parameter error, please check the parameter. " +
                                "'children' is not a list.")
                
        return numbered_list_block
    
    # FC: [Block dict] to do list
    @staticmethod
    def get_to_do_list_block(text:str, checked:bool=False, color:str="default", **kwargs) ->dict:
        """
        Get to do list block dict
        
        Parameters:
            text:                   (str)       - Target block text
            checked:                (bool)      - Target block checked
            kwargs:                 (dict)      - Target block data
            
        Returns:
            to_do_list_block:       (dict)      - Target block dict
        """
        if not Dict_Gadget.check_text_color(color):
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter. " +
                            "'color' is not in TEXT_COLOR_LIST. \n" +
                            f"TEXT_COLOR_LIST = {TEXT_COLOR_LIST}")
        
        type = 'to_do'
        to_do_list_text = {'text': {'content': text}}
        if 'style' in kwargs:
            to_do_list_text.update({'annotations':kwargs['style']})
        else:
            to_do_list_text.update({'annotations':Property_gadget.set_text_style(**kwargs)})
        
        to_do_list_block = {'type': type,
                            type: {
                                    'rich_text': [to_do_list_text],
                                    'checked': checked,
                                    'color': color,
                                  },
                            }
        
        if "children" in kwargs:
            if type(kwargs["children"]) is list:  # type: ignore "kwargs["children"] need be is list"
                to_do_list_block[type].update({'children': kwargs["children"]})
            else:
                raise Exception(f"{sys._getframe().f_code.co_name}: " +
                                "Parameter error, please check the parameter. " +
                                "'children' is not a list.")
                
        return to_do_list_block
    
    # FC: [Block dict] toggle list
    @staticmethod
    def get_toggle_list_block(text:str, color:str="default", **kwargs) ->dict:
        """
        Get toggle list block dict
        
        Parameters:
            text:                   (str)       - Target block text
            kwargs:                 (dict)      - Target block data
            
        Returns:
            toggle_list_block:      (dict)      - Target block dict
        """
        if not Dict_Gadget.check_text_color(color):
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter. " +
                            "'color' is not in TEXT_COLOR_LIST. \n" +
                            f"TEXT_COLOR_LIST = {TEXT_COLOR_LIST}")
        
        type = 'toggle'
        toggle_list_text = {'text': {'content': text}}
        if 'style' in kwargs:
            toggle_list_text.update({'annotations':kwargs['style']})
        else:
            toggle_list_text.update({'annotations':Property_gadget.set_text_style(**kwargs)})
        
        toggle_list_block = {'type': type,
                            type: {
                                    'rich_text': [toggle_list_text],
                                    'color': color,
                                  },
                            }
        
        if "children" in kwargs:
            if type(kwargs["children"]) is list:   # type: ignore "kwargs["children"] need be is list"
                toggle_list_block[type].update({'children': kwargs["children"]})
            else:
                raise Exception(f"{sys._getframe().f_code.co_name}: " +
                                "Parameter error, please check the parameter. " +
                                "'children' is not a list.")
                
        return toggle_list_block
    
    # FC: [Block dict] code page
    @staticmethod
    def get_code_page_block(text:str, language:str="plaintext") ->dict:
        """
        Get code page block dict
        
        Parameters:
            text:                   (str)       - Target block text
            language:               (str)       - Target block language
            
        Returns:
            code_page_block:        (dict)      - Target block dict
        """
        if not Dict_Gadget.check_code_language(language):
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter. " +
                            "'language' is not in CODE_LANGUAGE_LIST. \n" +
                            f"CODE_LANGUAGE_LIST = {CODE_LANGUAGE_LIST}")
            
        type = 'code'
        code_page_block = {'type': type,
                            type: {
                                    'type': "text",
                                    'text': {'content': text
                                            },
                                    'language': language,
                                  },
                            }
        
        return code_page_block
        
    # FC: [Block dict] child page
    # TODO: untest
    # WARNING: To create or update child_page type blocks, use the "Create Page" and the "Update page" endpoint.
    @staticmethod
    def get_child_page_block(title:str) ->dict:
        """
        Get child page block dict
        
        Parameters:
            child_page_id:          (str)       - Target block child page id
            title:                  (str)       - Target block title
            kwargs:                 (dict)      - Target block data
            
        Returns:
            child_page_block:       (dict)      - Target block dict
        """
        type = 'child_page'
        child_page_block = {'type': type,
                            type: {
                                    'title': title,
                                  },
                            }
        return child_page_block
    
    # FC: [Block dict] child database
    # WARNING: To create or update child_database type blocks, use the "Create database" and the "Update database" endpoint.
    @staticmethod
    def get_child_database_block(title:str) ->dict:
        """
        Get child database block dict
        
        Parameters:
            child_database_id:     (str)       - Target block child database id
            title:                  (str)       - Target block title
            kwargs:                 (dict)      - Target block data
            
        Returns:
            child_database_block:  (dict)      - Target block dict
        """
        type = 'child_database'
        child_database_block = {'type': type,
                                type: {
                                        'title': title,
                                    },
                                }
        return child_database_block
    
    # FC: [Block dict] embed block
    @staticmethod
    def get_embed_block(url:str) ->dict:
        """
        Get embed block dict
        
        Parameters:
            url:                    (str)       - Target block url
            
        Returns:
            embed_block:            (dict)      - Target block dict
        """
        type = 'embed'
        embed_block = {'type': type,
                        type: {
                                'url': url,
                                },
                        }
        return embed_block
    
    # FC: [Block dict] media block
    @staticmethod
    def get_media_block(url:str, type:str) ->dict:
        """
        Get image block dict
        
        Parameters:
            image_url:              (str)       - Target block image url
            
        Returns:
            image_block:            (dict)      - Target block dict
            
        Raises:
            Exception:              (Exception) - Parameter error
        """
        if not Dict_Gadget.check_media_type(type):
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter. " +
                            "'type' is not in MEDIA_TYPE_LIST. \n" +
                            f"MEDIA_TYPE_LIST = {MEDIA_TYPE_LIST}")
        media_block = {'type': type,
                        type: {
                                'type': "external",
                                'external': {'url':  url
                                            }
                                },
                        }
        return media_block
    
    # FC: [Block dict] bookmark block
    @staticmethod
    def get_bookmark_block(url:str, title:str) ->dict:
        """
        Get bookmark block dict
        
        Parameters:
            url:                    (str)       - Target block url
            title:                  (str)       - Target block title
            
        Returns:
            bookmark_block:         (dict)      - Target block dict
        """
        type = 'bookmark'
        bookmark_block = {'type': type,
                        type: {
                                'url': url,
                                },
                        }
        return bookmark_block
    
    # FC: [Block dict] equation block
    @staticmethod
    def get_equation_block(text:str) ->dict:
        """
        Get equation block dict
        
        Parameters:
            text:                   (str)       - Target block text
            
        Returns:
            equation_block:        (dict)      - Target block dict
            
        Raises:
            Exception:              (Exception) - Parameter error
        """
        type = 'equation'
        equation_block = {'type': type,
                        type: {
                                'expression': text,
                                },
                        }
        return equation_block
    
    # FC: [Block dict] divider block
    @staticmethod
    def get_divider_block() ->dict:
        """
        Get divider block dict
        
        Returns:
            divider_block:         (dict)      - Target block dict
            
        Raises:
            Exception:              (Exception) - Parameter error
        """
        type = 'divider'
        divider_block = {'type': type,
                        type: {},
                        }
        return divider_block
    
    # FC: [Block dict] table of contents block
    @staticmethod
    def get_table_of_contents_block(color:str="default") ->dict:
        """
        Get table of contents block dict
        
        Parameters:
            color:                  (str)       - Target block color
            
        Returns:
            table_of_contents_block:(dict)      - Target block dict
            
        Raises:
            Exception:              (Exception) - Parameter error
        """
        if not Dict_Gadget.check_text_color(color):
            raise Exception(f"{sys._getframe().f_code.co_name}: " +
                            "Parameter error, please check the parameter. " +
                            "'color' is not in COLOR_LIST. \n" +
                            f"COLOR_LIST = {TEXT_COLOR_LIST}")
        type = 'table_of_contents'
        table_of_contents_block = {'type': type,
                                    type: {
                                            'color': color,
                                            },
                                    }
        return table_of_contents_block
    
    # FC: [Block dict] breadcrumb block
    @staticmethod
    def get_breadcrumb_block() ->dict:
        """
        Get breadcrumb block dict
        
        Returns:
            breadcrumb_block:      (dict)      - Target block dict
        """
        type = 'breadcrumb'
        breadcrumb_block = {'type': type,
                            type: {},
                            }
        return breadcrumb_block
    
    # FC: [Block dict] column_list block
    @staticmethod
    def get_column_list_block(children:list=[]) ->dict:
        """
        Get column_list block dict
        
        Parameters:
            children:               (list)      - Target block children
        
        Returns:
            column_list_block:      (dict)      - Target block dict
        """
        type = 'column_list'
        column_list_block = {'type': type,
                            type: {'children': children},
                            }
        return column_list_block
    
    # FC: [Block dict] column block
    @staticmethod
    def get_column_block(children:list=[]) ->dict:
        """
        Get column block dict
        
        Parameters:
            children:               (list)      - Target block children
            
        Returns:
            column_block:           (dict)      - Target block dict
        """
        type = 'column'
        column_block = {'type': type,
                        type: {'children': children},
                        }
        return column_block
    
    # FC: [Block dict] link preview block
    @staticmethod
    def get_link_preview_block(url:str) ->dict:
        """
        Get link preview block dict
        
        Parameters:
            url:                  (str)       - Target block url
        
        Returns:
            link_preview_block:   (dict)      - Target block dict
        
        """
        type = 'link_preview'
        link_preview_block = {'type': type,
                            type: {'url': url},
                            }
        return link_preview_block
    
    # FC: [Block dict] template block
    @staticmethod
    def get_template_block(template_title:str, **kwargs) ->dict:
        """
        Get template block dict
        
        Returns:
            template_block:        (dict)      - Target block dict
            
        Raises:
            Exception:             (Exception) - Parameter error
        """
        type = 'template'
        
        template_title_text = {'text': {'content': template_title}}
        if 'style' in kwargs:
            template_title_text.update({'annotations':kwargs['style']})
        else:
            template_title_text.update({'annotations':Property_gadget.set_text_style(**kwargs)})
            
        template_block = {'type': type,
                          type:{'rich_text': [template_title_text]},
                        }
        if "children" in kwargs:
            if type(kwargs["children"]) is list:  # type: ignore "kwargs["children"] need be is list"
                template_block[type].update({'children': kwargs["children"]})
            else:
                raise Exception(f"{sys._getframe().f_code.co_name}: " +
                                "Parameter error, please check the parameter. " +
                                "'children' is not a list.")
            
        return template_block
    
    # FC: [Block dict] link to page block
    @staticmethod
    def get_link_to_page_block(page_id:str) ->dict:
        """
        Get link to page block dict
        
        Parameters:
            page_id:               (str)       - Target block page id
        
        Returns:
            link_to_page_block:    (dict)      - Target block dict
        """
        type = 'link_to_page'
        link_to_page_block = {'type': type,
                              type: {
                                  'type': "page_id",
                                  'page_id': page_id},
                            }
        return link_to_page_block
    
    # FC: [Block dict] link to database block
    @staticmethod
    def get_link_to_database_block(database_id:str) ->dict:
        """
        Get link to page block dict
        
        Parameters:
            page_id:               (str)       - Target block page id
        
        Returns:
            link_to_page_block:    (dict)      - Target block dict
        """
        type = 'link_to_page'
        link_to_page_block = {'type': type,
                              type: {
                                  'type': "database_id",
                                  'page_id': database_id},
                            }
        return link_to_page_block
        
    # FC: [Block dict] synced block
    @staticmethod
    def get_synced_block(children:list=[]) ->dict:
        """
        Get synced block dict
        
        Parameters:
            children:               (list)      - Target block children
        
        Returns:
            synced_block:           (dict)      - Target block dict
        """
        type = 'synced_block'
        synced_block = {'type': type,
                        type: {
                            'children': children},
                        }
        return synced_block
    
    # FC: [Block dict] synced from block
    @staticmethod
    def get_synced_from_block(block_id:str) ->dict:
        """
        Get synced from block dict
        
        Parameters:
            children:               (list)      - Target block children
        
        Returns:
            synced_from_block:      (dict)      - Target block dict
        """
        type ='synced_block'
        synced_from_block = {'type': type,
                            type: {'synced_from': {"block_id": block_id}
                                    },
                            }
        return synced_from_block

class Dict_Gadget:
    Property = Property_gadget()
    Page = Page_gadget()
    Database = Database_gadget()
    Block = Block_gadget()
    
    #--------------------------[Object]--------------------------#
    # FC [Gadget]: Get Icon object
    @staticmethod
    def get_icon_dict(emoji:str) ->dict:
        return Icon(emoji=emoji).asdict()
    
    #--------------------------[Time]--------------------------#
    # FC [Gadget-Time]: Set time str
    @staticmethod
    def set_time(date:str, time:str, timezone:str) ->str:
        """
        Set time
        
        Parameters:
            date:                   (str)           - Date ex: 2021-01-01
            time:                   (str)           - Time ex: 00:00:00
            timezone:               (str)           - Target timezone ex: Asia/Tokyo
        
        """
        tz = pytz.timezone(timezone)
        tz_offset = str(datetime.datetime.now(tz))[-6:]
        return date + "T" + time + ".000" + tz_offset
    
    # FC [Gadget-Time]: Get now time str
    @staticmethod
    def get_now_time(timezone:str) ->str:
        """
        Get now time

        Returns:
            time:                   (str)           - Time ex: Asia/Tokyo
        """
        tz = pytz.timezone(timezone)
        tz_offset = str(datetime.datetime.now(tz))[-6:]
        now_time = time.strftime("%Y-%m-%dT%H:%M:%S.000", time.localtime(time.time()))
        return now_time + tz_offset
    
    # FC [Gadget-Time]: Timezone change
    @staticmethod
    def timezone_change(utc_time:str, timezone:str) ->str:
        """
        UTC time change to timezone time

        Parameters:
            utc_time:               (str)           - UTC time
            timezone:               (str)           - Target timezone
        
        returns:
            timezone_time:          (str)           - Timezone time
        """
        tz = pytz.timezone(timezone)
        date_temp = "%Y-%m-%d"; time_temp = "%H:%M:%S"
        time_utc = datetime.datetime.fromisoformat(utc_time.replace('Z', '+00:00'))
        time_tz = time_utc.astimezone(tz)
        return time_tz.strftime(date_temp +" "+ time_temp)

    #--------------------------[URL]--------------------------#
    # FC [Gadget-URL]: Notion Url to short_url
    @staticmethod
    def get_notion_short_url(url:str) ->str:
        """
        Get notion short url
        
        Parameters:
            url:                    (str)           - Target url

        Returns:
            result:                 (str)           - Result
        """
        return '/' + get_id(url).replace('-','')
    
    #--------------------------[Check]--------------------------#
    # FC [Gadget-Check]: Is Notion url
    @staticmethod
    def is_notion_url(url:str) ->bool:
        """
        Check if the url is a notion url
        
        Parameters:
            url:                   (str)          - Target url

        Returns:
            result:                 (bool)          - Result
        """
        if url is None:
            return False
        return url[:22] == "https://www.notion.so/"

    # FC: [Gadget-Check] text color
    @staticmethod
    def check_text_color(color:str) ->bool:
        """
        Check text color
        
        Parameters:
            color:                  (str)           - Target color

        Returns:
            result:                 (bool)          - Result
        """
        return color in TEXT_COLOR_LIST
    
    # FC: [Gadget-Check] rollup function
    @staticmethod
    def check_rollup_function(function:str) ->bool:
        """
        Check rollup function
        
        Parameters:
            function:               (str)           - Target function

        Returns:
            result:                 (bool)          - Result
        """
        return function in ROLLUP_FUNCTION_LIST
    
    # FC: [Gadget-Check] block type
    @staticmethod
    def check_block_type(type:str) ->bool:
        """
        Check block type
        
        Parameters:
            type:                   (str)           - Target type

        Returns:
            result:                 (bool)          - Result
        """
        return type in BLOCK_TYPE_LIST
    
    # FC: [Gadget-Check] code language
    @staticmethod
    def check_code_language(language:str) ->bool:
        """
        Check code language
        
        Parameters:
            language:               (str)           - Target language

        Returns:
            result:                 (bool)          - Result
        """
        return language in CODE_LANGUAGE_LIST
    
    # FC: [Gadget-Check] media type
    @staticmethod
    def check_media_type(type:str) ->bool:
        """
        Check media type
        
        Parameters:
            type:                   (str)           - Target type

        Returns:
            result:                 (bool)          - Result
        """
        return type in MEDIA_TYPE_LIST
    
    # FC: [Gadget-Check] is page dict
    @staticmethod
    def is_page(page:dict) ->bool:
        """
        Check if the page is a page
        
        Parameters:
            page:                   (dict)          - Target page

        Returns:
            result:                 (bool)          - Result
        """
        return is_full_page(page)
    
    # FC: [Gadget-Check] is database dict
    @staticmethod
    def is_database(database:dict) ->bool:
        """
        Check if the database is a database
        
        Parameters:
            database:               (dict)          - Target page

        Returns:
            result:                 (bool)          - Result
        """
        return is_full_database(database)
    
    # FC: [Gadget-Check] check dict is page or database
    @staticmethod
    def check_dict_type(dict:dict) ->str:
        """
        Check dict type
        
        Parameters:
            dict:                   (dict)          - Target page

        Returns:
            result:                 (str)           - Result page or database
        """
        if dict['object'] == 'page':
            return 'page'
        elif dict['object'] == 'database':
            return 'database'
        else:
            return 'none'



####################################################################################
# MIT License
#
# Copyright (c) 2023 bluewhitep
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
####################################################################################
from dataclasses import dataclass, field, asdict

from .CONTENTS import (
    ALL_PROPERTIES_TYPES,
    ROLLUP_FUNCTION_LIST,
    ROLLUP_TYPE_LIST,
    TEXT_COLOR_LIST,
    FORMAT_TYPES,
    CODE_LANGUAGE_LIST,
    BLOCK_TYPE_LIST,
    BLOCK_CHILDREN_TYPES,
)

def clear_empty_id(Dict:dict) ->None:
    if 'id' in Dict and Dict['id'] == '':
        Dict.pop('id')

def clear_empty_name(Dict:dict) ->None:
    if 'name' in Dict and Dict['name'] == '':
        Dict.pop('name')

def check_color(color):
    if color not in TEXT_COLOR_LIST:
        raise ValueError(f"Color {color} is not in {TEXT_COLOR_LIST}")

# FC: [Base] object class generic method
@dataclass
class BaseMethod:
    def __post_init__(self) -> None:
        self.Dict = self.asdict()
                              
    def asdict(self) -> dict: 
        Dict = asdict(self) 
        clear_empty_id(Dict)
        clear_empty_name(Dict)
        return Dict

    def update(self):
        self.Dict = self.asdict()
        
# FC: [Base] color
@dataclass
class ColorMethod(BaseMethod):
    color: str = 'default'
    
#--------------------------[L1 Base: Database-property type ]---------------------#  
# FC: [L1 Base T] Number
@dataclass
class Number(BaseMethod): 
    format: str = 'number'
    
    def __post_init__(self) -> None:
        assert self.format in FORMAT_TYPES, f"Format {self.format} is not in {FORMAT_TYPES}"
        super().__post_init__()
        
# FC: [L1 Base T] formula
@dataclass
class Formula(BaseMethod): 
    expression: str = field(default_factory=str)

# FC: [L1 Base T/I] option base
@dataclass
class Option(BaseMethod): 
    id: str | None = field(default_factory=str)
    name: str = field(default_factory=str)
    color: str = field(default_factory=str)
    
# FC: [L1 Base T] group base
@dataclass
class Group(BaseMethod): 
    id: str = field(default_factory=str)
    name: str = field(default_factory=str)
    color: str = field(default_factory=str)
    option_ids: list[str] = field(default_factory=list)
    
# FC: [L1 Base T] relation synced
@dataclass
class RelationSynced(BaseMethod): 
    synced_property_id: str = field(default_factory=str)
    synced_property_name: str = field(default_factory=str)

# FC: [L1 Base T] rollup
@dataclass
class Rollup(BaseMethod):
    function: str = "show_original"
    relation_property_name: str = field(default_factory=str)
    relation_property_id: str = field(default_factory=str)
    rollup_property_name: str = field(default_factory=str)
    rollup_property_id: str = field(default_factory=str)

#--------------------------[L1 Base: property item]---------------------#
# FC: [L1 Base I] Url
@dataclass
class Url(BaseMethod): 
    url: str = field(default_factory=str)
    
# FC: [L1 Base I] Icon
@dataclass
class Icon(BaseMethod): 
    type: str = 'emoji'
    emoji: str = field(default_factory=str)

# FC: [L1 Base I] parent
@dataclass
class Parent(BaseMethod): 
    type: str = field(default_factory=str)
    page_id: str = field(default_factory=str)
    database_id: str = field(default_factory=str)
    
    def asdict(self) -> dict:
        Dict = super().asdict()
        if Dict['type'] == 'page_id':
            Dict.pop('database_id')
        elif Dict['type'] == 'database_id':
            Dict.pop('page_id')
        return Dict
    
# FC: [L1 Base I] Email
@dataclass
class Email(BaseMethod): 
    email: str = field(default_factory=str) 
 
# FC: [L1 Base I] Date
@dataclass
class Date(BaseMethod): 
    start: str = field(default_factory=str)
    end: str | None = field(default_factory=str)
    time_zone: None = None
    
# FC: [L1 Base I] formula value
@dataclass
class FormulaValue(BaseMethod): 
    """
    FormulaValue
    
    Parameters:
        type        (str):      - The type of the formula result. [Default: "string"]
                                - Possible values are "string", "number", "boolean", and "date".
    """
    type: str = 'string'
    
    # formula result
    string: str | None = None
    number: float | None = None
    boolean: bool = False
    date: Date | None = None
    
    def __post_init__(self) -> None:
        if type(self.date) == dict:
            self.date = Date(**self.date) # type: ignore
        super().__post_init__()
    
    def __zip(self) -> dict:
        return dict(zip(["string", "number", "boolean", "date"],
                        [self.string, self.number, self.boolean, self.date]))
    
    def asdict(self) -> dict: 
        _zip_value = self.__zip()
        Dict = {'type': self.type,
                self.type: _zip_value[self.type].asdict() if type(_zip_value[self.type]) == Date else _zip_value[self.type] 
            }
        return Dict 

# FC: [L1 Base I] rollup item
@dataclass
class RollupItem(BaseMethod): 
    """
    RollupItem
    
    Parameters:
        type        (str):      - The type of the rollup result. [Default: "array"]
                                  Possible values are "number", "date", "array", "unsupported" and "incomplete"
        function    (str):      - The function used to calculate the rollup result. [Default: "show_original"]
                                  Possible values include: "count", "count_values", "empty", "not_empty", "unique", 
                                    "show_unique", "percent_empty", "percent_not_empty", "sum", "average", "median", 
                                    "min", "max", "range", "earliest_date", "latest_date", "date_range", "checked", 
                                    "unchecked", "percent_checked", "percent_unchecked", "count_per_group", "percent_per_group", 
                                    "show_original"
    
    Raise
        ValueError:   - If the "function" is not in the ROLLUP_FUNCTION_LIST
        ValueError:   - If the "type" is not in the ROLLUP_TYPE_LIST
    """
    function: str = "show_original"
        #count, count_values, empty, not_empty, unique, show_unique, percent_empty, percent_not_empty, sum, average, median, min, max, range, earliest_date, latest_date, date_range, checked, unchecked, percent_checked, percent_unchecked, count_per_group, percent_per_group, show_original
    type: str = "array"
        # "number", "date", "array", "unsupported" and "incomplete"
    
    number: int = 0
    date: Date | None = None
    array: list[dict] = field(default_factory=list) # [{type:value}]
    incomplete: dict = field(default_factory=dict)
      
    def __post_init__(self) -> None:
        if type(self.date) == dict:
            self.date = Date(**self.date) # type: ignore
        
        assert self.function in ROLLUP_FUNCTION_LIST, f"Function {self.function} is not in the {ROLLUP_FUNCTION_LIST}"
        assert self.type in ROLLUP_TYPE_LIST, f"Type {self.type} is not in the {ROLLUP_TYPE_LIST}"
        super().__post_init__()
    
    def __zip(self) -> dict:
        return dict(zip(["number", "date", "array", "incomplete"],
                        [self.number, self.date, self.array, self.incomplete])) 
        
    def asdict(self) -> dict: 
        _zip_value = self.__zip()
        Dict = {'type': self.type,
                'function': self.function,
                self.type: _zip_value[self.type].asdict() if type(_zip_value[self.type]) == Date else _zip_value[self.type]
            }
        return Dict

#--------------------------[User]---------------------#
# FC: [User] bot owner
@dataclass
class BotOwner(BaseMethod): 
    type: str = 'workspace'
    workspace: bool = True
    
# FC: [User] bot
@dataclass
class BotBase(BaseMethod): 
    owner: BotOwner = field(default_factory=BotOwner)
    workspace_name: str = field(default_factory=str)
    
    def __post_init__(self) -> None:
        if type(self.owner) == dict:
            self.owner = BotOwner(**self.owner) # type: ignore
        super().__post_init__()
       
# FC: [User] short object
@dataclass
class UserShort(BaseMethod): 
    id: str = field(default_factory=str)
    object: str = field(default_factory=str)

# FC: [User] user object 
@dataclass
class User(BaseMethod):
    avatar_url: str | None = field(default_factory=str)
    id: str = field(default_factory=str)
    name: str = field(default_factory=str)
    object: str = 'user'
    person: Email = field(default_factory=Email)
    type: str = 'person'
    
    def __post_init__(self) -> None:
        if type(self.person) == dict:
            self.person = Email(**self.person) # type: ignore
        super().__post_init__()
        
# FC: [User] bot object
@dataclass
class Bot(BaseMethod):
    avatar_url: str | None = None
    bot: BotBase = field(default_factory=BotBase)
    id: str = field(default_factory=str)
    name: str = field(default_factory=str)
    object: str = 'user'
    type: str = 'bot'
    
    def __post_init__(self) -> None:
        if type(self.bot) == dict:
            self.bot = BotBase(**self.bot) # type: ignore
        super().__post_init__()

#--------------------------[File]---------------------#
# FC: [file] upload
@dataclass
class Upload(BaseMethod): 
    url: str = field(default_factory=str)
    expiry_time: str = field(default_factory=str)
    
# FC: [file] object
@dataclass
class FileUpload(BaseMethod): 
    type: str = 'file'
    file: Upload = field(default_factory=Upload)
    name: str = field(default_factory=str)
    
    def __post_init__(self) -> None:
        if type(self.file) == dict:
            self.file = Upload(**self.file) # type: ignore
        super().__post_init__()

# FC: [file] object
@dataclass
class FileLink(BaseMethod): 
    type: str = 'external'
    external: Url = field(default_factory=Url)
    name: str = field(default_factory=str)
    
    def __post_init__(self) -> None:
        self.external = Url(**self.external) if type(self.external) == dict else self.external # type: ignore
        super().__post_init__()

#--------------------------[Rich Text]---------------------#
# FC: [Text] content
@dataclass
class TextContent(BaseMethod): 
    content: str = field(default_factory=str)
    link: Url | None = None
    
    def __post_init__(self) -> None:
        if type(self.link) == dict:
            self.link = Url(**self.link) # type: ignore
        super().__post_init__()
    
    def text_dict(self, name:str="text"): 
        return {name: self.asdict()}
        
# FC: [Text] style
@dataclass
class TextStyle(BaseMethod):
    bold: bool = False
    italic: bool = False
    strikethrough: bool = False
    underline: bool = False
    code: bool = False
    color: str ="default"
    
    def __post_init__(self) -> None:
        check_color(self.color)
        super().__post_init__()
    
# FC: [Text] object
@dataclass
class RichText(BaseMethod): 
    type: str = "text"
    text: TextContent = field(default_factory=TextContent)
    annotations: TextStyle = field(default_factory=TextStyle)
    href: str | None = None
    plain_text: str | None = None
    
    def __post_init__(self) -> None:
        self.text = TextContent(**self.text) if type(self.text) == dict else self.text # type: ignore
        self.annotations = TextStyle(**self.annotations) if type(self.annotations) == dict else self.annotations # type: ignore
        self.href = self.text.link.url if self.text.link is not None else None
        self.plain_text = self.text.content
        super().__post_init__()

#--------------------------[L2 Base]---------------------#
# FC: [L2] title
@dataclass
class Title(BaseMethod): 
    title: list[RichText] = field(default_factory=list)

    def __post_init__(self) -> None:
        new_title = []
        for _,rich_text in enumerate(self.title):
            if type(rich_text) == dict:
                new_title.append(RichText(**rich_text)) # type: ignore
        self.title = new_title
        super().__post_init__()

# FC: [L2 Base] options
@dataclass
class Options(BaseMethod): 
    options: list[Option] = field(default_factory=list)

    def __post_init__(self) -> None:
        new_options = []
        for _,option in enumerate(self.options):
            if type(option) == dict:
                new_options.append(Option(**option)) # type: ignore
            else:
                new_options.append(option)
        self.options = new_options
        super().__post_init__()

    def get_options_list(self) -> list[dict]:
        return [_.Dict for _ in self.options]
    
# FC: [L2 Base] groups
@dataclass
class Groups(BaseMethod): 
    groups: list[Group] = field(default_factory=list)
    
    def __post_init__(self) -> None:
        new_groups = []
        for _, group in enumerate(self.groups):
            if type(group) == dict:
                new_groups.append(Group(**group)) # type: ignore
            else:
                new_groups.append(group)
        self.groups = new_groups
        super().__post_init__()

# FC: [L2 Base] relation single
@dataclass
class RelationSingle(BaseMethod): 
    database_id: str = field(default_factory=str)
    single_property: dict = field(default_factory=dict)
    type: str = 'single_property'

# FC: [L2 Base] relation dual
@dataclass
class RelationDual(BaseMethod): 
    type: str = 'dual_property'
    database_id: str = field(default_factory=str)
    dual_property: RelationSynced = field(default_factory=RelationSynced)

    def __post_init__(self) -> None:
        if type(self.dual_property) == dict:
            self.dual_property = RelationSynced(**self.dual_property) # type: ignore
        super().__post_init__()

# FC: [L2 Base] status
@dataclass
class Status(BaseMethod): 
    groups: list[Group] = field(default_factory=list)
    options: list[Option] = field(default_factory=list)
    
    def __post_init__(self) -> None:
        new_groups = []
        for _, group in enumerate(self.groups):
            if type(group) == dict:
                new_groups.append(Group(**group)) # type: ignore
            else:
                new_groups.append(group)
        new_options = []
        for _,option in enumerate(self.options):
            if type(option) == dict:
                new_options.append(Option(**option)) # type: ignore
            else:
                new_options.append(option)
        self.options = new_options
        super().__post_init__()

#--------------------------[Request return]---------------------#
# FC: [Other] Request info
@dataclass
class RequestInfo(BaseMethod): 
    archived: bool = False
    cover: FileLink | None = None
    created_by: UserShort = field(default_factory=UserShort)
    created_time: str = field(default_factory=str)
    icon: Icon | None = None
    id: str = field(default_factory=str)
    last_edited_by: UserShort = field(default_factory=UserShort)
    last_edited_time: str = field(default_factory=str)
    parent: Parent = field(default_factory=Parent)
    url: Url = field(default_factory=Url)
    object: str = field(default_factory=str)
    
    def __post_init__(self) -> None:
        if type(self.created_by) == dict:
            self.created_by = UserShort(**self.created_by) # type: ignore
        if type(self.last_edited_by) == dict:
            self.last_edited_by = UserShort(**self.last_edited_by) # type: ignore
        if type(self.parent) == dict:
            self.parent = Parent(**self.parent) # type: ignore
        if type(self.url) == dict:
            self.url = Url(**self.url) # type: ignore
        if type(self.icon) == dict:
            self.icon = Icon(**self.icon) # type: ignore
        if type(self.cover) == dict:
            self.cover = FileLink(**self.cover) # type: ignore
        super().__post_init__()

    def asdict(self) -> dict: 
        Dict = super().asdict()
        if Dict['cover'] is not None and Dict['cover']['name'] == '':
            Dict['cover'].pop('name')
            
        if Dict['parent']['type'] == 'database_id':
            Dict['parent'].pop('page_id')
        elif Dict['parent']['type'] == 'page_id':
            Dict['parent'].pop('database_id')
            
        return Dict

#--------------------------[Block Base]---------------------#

# FC: [Block Base] Retrieve request info base
@dataclass
class BlockBase(BaseMethod):
    archived: bool = False
    created_by: UserShort = field(default_factory=UserShort)
    created_time: str = field(default_factory=str)
    has_children: bool = False
    id: str = field(default_factory=str)
    last_edited_by: UserShort = field(default_factory=UserShort)
    last_edited_time: str = field(default_factory=str)
    object: str = 'block'
    parent: Parent = field(default_factory=Parent)
    
    def __post_init__(self) -> None:
        if type(self.parent) == dict:
            self.parent = Parent(**self.parent) # type: ignore
        if type(self.created_by) == dict:
            self.created_by = UserShort(**self.created_by) # type: ignore
        if type(self.last_edited_by) == dict:
            self.last_edited_by = UserShort(**self.last_edited_by) # type: ignore
        super().__post_init__()
        
    def asdict(self) -> dict:
        Dict = super().asdict()
        
        Dict['parent'] = self.parent if type(self.parent) == dict else self.parent.asdict()
        Dict['created_by'] = self.created_by if type(self.created_by) == dict else self.created_by.asdict()
        Dict['last_edited_by'] = self.last_edited_by if type(self.last_edited_by) == dict else self.last_edited_by.asdict()
            
        return Dict

# FC: [Block Base] Block ID
@dataclass
class BlockID(BaseMethod):
    block_id: str = field(default_factory=str)

# FC: [Blocl Base] paragraph base
@dataclass
class ParagraphBase(BaseMethod):
    """
    Raises:
        ValueError:  If the "color" is not in the TEXT_COLOR_LIST
    """
    rich_text: list[RichText] = field(default_factory=list)
    color: str = "default"
    
    def __post_init__(self) -> None:
        if self.color not in TEXT_COLOR_LIST:
            raise ValueError(f"Text color {self.color} is not in the TEXT_COLOR_LIST")
        self.rich_text = [RichText(**text) if type(text) == dict else text # type: ignore
                            for text in self.rich_text]
        super().__post_init__()

# FC: [Blocl Base] paragraph text base
@dataclass
class Paragraph(ParagraphBase):
    pass

# FC: [Block Base] heading_1
@dataclass
class Heading(ParagraphBase):
    is_toggleable: bool = False

# FC: [Block Base] callout
@dataclass
class Callout(ParagraphBase):
    icon: Icon = field(default_factory=Icon)
    
    def __post_init__(self) -> None: 
        self.icon = Icon(**self.icon) if type(self.icon) == dict else self.icon # type: ignore
        super().__post_init__()

# FC: [Block Base] quote
@dataclass
class Quote(ParagraphBase):
    pass

# FC: [Block Base] bulleted_list_item
@dataclass
class BulletedListItem(ParagraphBase):
    pass

# FC: [Block Base] number_list_item
@dataclass
class NumberedListItem(ParagraphBase):
    pass

# FC: [Block Base] to_do
@dataclass
class ToDo(ParagraphBase):
    checked: bool = False
    
# FC: [Block Base] toggle
@dataclass
class Toggle(ParagraphBase):
    pass
    
# FC: [Block Base] code
@dataclass
class Code(BaseMethod):
    """
    Raises:
        ValueError:  If the "color" is not in the TEXT_COLOR_LIST
    """
    rich_text: list[RichText] = field(default_factory=list)
    caption: list[RichText] = field(default_factory=list)
    language: str = "python"
    
    def __post_init__(self) -> None:
        if self.language not in CODE_LANGUAGE_LIST:
            raise ValueError(f"Text color {self.language} is not in the CODE_LANGUAGE_LIST")
        self.rich_text = [RichText(**text) if type(text) == dict else text # type: ignore
                            for text in self.rich_text]
        self.caption = [RichText(**text) if type(text) == dict else text # type: ignore
                            for text in self.caption]
        super().__post_init__()

# FC: [Block Base] child_page
@dataclass
class ChildPage(BaseMethod):
    title: str = field(default_factory=str)

# FC: [Block Base] child_database
@dataclass
class ChildDatabase(BaseMethod):
    title: str = field(default_factory=str)

# FC: [Block Base] expression
@dataclass
class Expression(BaseMethod):
    expression: str = field(default_factory=str)

# FC: [Block Base] template
@dataclass
class Template(BaseMethod):
    rich_text: list[RichText] = field(default_factory=list)
    
    def __post_init__(self) -> None:
        self.rich_text = [RichText(**text) if type(text) == dict else text # type: ignore
                            for text in self.rich_text]
        super().__post_init__()

# FC: [Block Base] link_to
@dataclass
class LinkTo(Parent):
    pass

# FC: [Block Base] synced block
@dataclass
class SyncedBlock(BaseMethod):
    synced_from: BlockID | None = None

    def __post_init__(self) -> None:
        if self.synced_from is not None:
            self.synced_from = BlockID(**self.synced_from) if type(self.synced_from) == dict else self.synced_from # type: ignore
        super().__post_init__()
    
# FC: [Block Base] table
@dataclass
class Table(BaseMethod):
    table_width: int = field(default_factory=int)
    has_column_header: bool = False
    has_row_header: bool = False
    
# FC: [Block Base] table row
@dataclass
class TableRow(BaseMethod):
    cells: list[list[RichText]] = field(default_factory=list)
    
    def __post_init__(self) -> None:
        self.cells = [[RichText(**text) if len(text) != 0 and type(text) == dict  else text # type: ignore
                       for text in cell] for cell in self.cells]

#--------------------------[Property]---------------------#
# FC: [L2 Base] property type
@dataclass
class PropertyType(BaseMethod): 
    id: str = field(default_factory=str)
    name: str = field(default_factory=str)
    type: str = 'title'
    
    # NOTE: Property type value
    # Because the value of key is type
    checkbox: dict = field(default_factory=dict)
    created_by: dict = field(default_factory=dict)
    created_time: dict = field(default_factory=dict)
    date: dict = field(default_factory=dict)
    email: dict = field(default_factory=dict)
    files: dict = field(default_factory=dict)
    formula: Formula = field(default_factory=Formula)
    last_edited_by: dict = field(default_factory=dict)
    last_edited_time: dict = field(default_factory=dict)
    multi_select: Options = field(default_factory=Options)
    title: dict | list[RichText] = field(default_factory=dict)
    number: Number = field(default_factory=Number)
    people: dict = field(default_factory=dict)
    phone_number: dict = field(default_factory=dict)
    relation: RelationSingle | RelationDual = field(default_factory=RelationSingle)
    select: Options = field(default_factory=Options)
    status: Status | dict = field(default_factory=dict)
    rich_text: dict = field(default_factory=dict)
    url:dict = field(default_factory=dict)
    rollup: Rollup = field(default_factory=Rollup)
        
    def __post_init__(self) -> None:
        if type(self.formula) == dict:
            self.formula = Formula(**self.formula) # type: ignore
        if type(self.relation) == dict:
            self.relation = RelationDual(**self.relation) if self.relation['type'] == 'dual_property' else RelationSingle(**self.relation)  # type: ignore
        if type(self.select) == dict:
            self.select = Options(**self.select)  # type: ignore
        if type(self.multi_select) == dict:
            self.multi_select = Options(**self.multi_select)  # type: ignore
        if type(self.status) == dict and self.status != {}:
            self.status = Status(**self.status)  # type: ignore
        if type(self.number) == dict:
            self.number = Number(**self.number)  # type: ignore
        if type(self.rollup) == dict:
            self.rollup = Rollup(**self.rollup)  # type: ignore
        if type(self.title) == list:
            new_title = []
            for _,rich_text in enumerate(self.title):
                if type(rich_text) == dict:
                    new_title.append(RichText(**rich_text))  # type: ignore
            self.title = new_title
        super().__post_init__()
    
    def __zip(self) -> dict: 
        return dict(zip(ALL_PROPERTIES_TYPES,
                        [self.title, self.rich_text, self.number, self.select, 
                        self.multi_select, self.status, self.date, self.people, 
                        self.files, self.checkbox, self.url, self.email, 
                        self.phone_number, self.relation, self.formula, 
                        self.rollup, self.created_time, self.created_by, 
                        self.last_edited_time, self.last_edited_by,
                        ]))
    
    def asdict(self) -> dict: 
        _zip_value = self.__zip()
        if type(_zip_value[self.type]) == dict:
            value = _zip_value[self.type]
        elif type(_zip_value[self.type]) == list:
            value = [_.asdict() for _ in _zip_value[self.type]]
        else:
            value = _zip_value[self.type].asdict()
        
        if 'options' in value or 'groups' in value:
            for _, options in value.items():  # type: ignore
                for option in options:
                    clear_empty_id(option)

        Dict = {
                "id": self.id,
                "name": self.name,
                "type": self.type,
                self.type: value,
        }
        clear_empty_id(Dict)
        return Dict

    def full_dict(self) -> dict: 
        return {self.name: self.asdict()}

    def label_dict(self) -> dict:
        return {'name': self.name,
                'type': self.type,
                'id': self.id,
                }
    
    def rename(self, new_name: str) -> None:
        self.name = new_name
        
# FC: [L2 Base] property Item
@dataclass
class PropertyItem(BaseMethod): 
    has_more: bool = False # NOTE: This is onlt for relation
    id: str = field(default_factory=str)
    type: str = field(default_factory=str)
    
    # NOTE: Property value
    title: list[RichText] = field(default_factory=list)
    rich_text: list[RichText] = field(default_factory=list)
    number: int | float | None = None
    select: Option | None = None
    multi_select: list[Option] = field(default_factory=list) 
    status: Option | None = None
    date: Date | None = None
    people: list[User] = field(default_factory=list)
    files: list[FileLink | FileUpload ] = field(default_factory=list)
    checkbox: bool = False
    url: str | None = None
    email: str | None = None
    phone_number: str | None = None
    relation: list[dict] = field(default_factory=list) # Example: [{'id': str}, ...]
    formula: FormulaValue = field(default_factory=FormulaValue)
    rollup: RollupItem = field(default_factory=RollupItem)
    created_by: User | Bot = field(default_factory=User)
    created_time: str = field(default_factory=str)
    last_edited_by: User | Bot = field(default_factory=User)
    last_edited_time: str = field(default_factory=str)
    
    def __post_init__(self) -> None:
        if type(self.created_by) == dict:
            self.created_by = Bot(**self.created_by) if self.created_by['type'] == 'bot' else User(**self.created_by) # type: ignore
        if type(self.date) == dict:
            self.date = Date(**self.date) # type: ignore
        if type(self.files) == list:
            new_files = []
            for file in self.files:
                if type(file) == dict and file['type'] == 'file': # type: ignore
                    new_files.append(FileUpload(**file)) # type: ignore
                elif type(file) == dict and file['type'] == 'external': # type: ignore
                    new_files.append(FileLink(**file)) # type: ignore
                else:
                    new_files.append(file)
            self.files = new_files
        if type(self.last_edited_by) == dict:
            self.last_edited_by = Bot(**self.last_edited_by) if self.last_edited_by['type'] == 'bot' else User(**self.last_edited_by) # type: ignore
        if type(self.formula) == dict:
            self.formula = FormulaValue(**self.formula) # type: ignore
        if type(self.multi_select) == list:
            new_multi_select = []
            for option in self.multi_select:
                if type(option) == dict:
                    new_multi_select.append(Option(**option)) # type: ignore
                else:
                    new_multi_select.append(option)
            self.multi_select = new_multi_select
        if type(self.title) == list:
            new_title = []
            for _,rich_text in enumerate(self.title):
                if type(rich_text) == dict:
                    new_title.append(RichText(**rich_text)) # type: ignore
                else:
                    new_title.append(rich_text)
            self.title = new_title
        if type(self.people) == list:
            new_people = []
            for user in self.people:
                if type(user) == dict:
                    new_people.append(User(**user)) # type: ignore
                else:
                    new_people.append(user)
            self.people = new_people
        if type(self.rollup) == dict:
            self.rollup = RollupItem(**self.rollup) # type: ignore
        if type(self.select) == dict:
            self.select = Option(**self.select) # type: ignore
        if type(self.status) == dict:
            self.status = Option(**self.status) # type: ignore
        if type(self.rich_text) == list:
            new_rich_text = []
            for _,rich_text in enumerate(self.rich_text):
                if type(rich_text) == dict:
                    new_rich_text.append(RichText(**rich_text)) # type: ignore
                else:
                    new_rich_text.append(rich_text)
            self.rich_text = new_rich_text
        super().__post_init__() 
        
    def __zip(self) -> dict: 
        return dict(zip(ALL_PROPERTIES_TYPES,
                        [self.title, self.rich_text, self.number, self.select, 
                        self.multi_select, self.status, self.date, self.people, 
                        self.files, self.checkbox, self.url, self.email, 
                        self.phone_number, self.relation, self.formula, 
                        self.rollup, self.created_time, self.created_by, 
                        self.last_edited_time, self.last_edited_by,
                        ]))

    def asdict(self) -> dict: 
        _zip_value = self.__zip()
        if (type(_zip_value[self.type]) == dict or
            type(_zip_value[self.type]) == str or
            type(_zip_value[self.type]) == int or
            type(_zip_value[self.type]) == float or
            type(_zip_value[self.type]) == bool or
            _zip_value[self.type] == None):
            value = _zip_value[self.type]
        elif type(_zip_value[self.type]) == list:
            value = [ _ if type(_) == dict else _.asdict()
                     for _ in _zip_value[self.type]]
        else:
            value = _zip_value[self.type].asdict()
        
        if (type(value) == dict and 
            ('options' in value or 'groups' in value)):
                for _, options in value.items(): # type: ignore
                    for option in options:
                        clear_empty_id(option)

        Dict = {
                "id": self.id,
                self.type: value,
                "type": self.type,
                }
        if self.type == 'relation':
            Dict['has_more'] = self.has_more
        
        clear_empty_id(Dict)
        return Dict

    def get_value(self):
        return self.Dict[self.type]

#--------------------------[Page]---------------------#
# FC: [Page] object
@dataclass
class Page(RequestInfo): 
    object: str = 'page'
    properties: dict[str, PropertyItem] = field(default_factory=dict)

    def __post_init__(self) -> None:
        new_properties = {}
        if type(self.properties) == dict:
            for name, value in self.properties.items():
                new_properties.update({name: PropertyItem(**value)}) # type: ignore
        self.properties = new_properties
        super().__post_init__()
    
    def asdict(self) -> dict: 
        Dict = super().asdict()
            
        new_properties = {}
        for name, value in self.properties.items():
            if type(value) != dict:
                value = value.asdict()
            new_properties.update({name: value}) 
        Dict['properties'] = new_properties
        
        return Dict
    
    def update_title(self, title:RichText):
        for _, item in self.properties.items():
            if item.type == 'title':
                item.title = [title]
                break
        self.update()
        
    def update_item(self, name:str, **kwargs):
        old_item = self.properties[name].Dict
        self.properties[name] = PropertyItem(id=old_item['id'],
                                             type=old_item['type'],
                                             **kwargs)
        self.update()
    
    def clear_item(self, name:str):
        old_item = self.properties[name].Dict
        self.properties[name] = PropertyItem(id=old_item['id'],
                                             type=old_item['type'])
        self.update()
        
    def get_item(self, name:str) ->PropertyItem:
        return self.properties[name]
    
    def get_items(self) ->dict[str, dict[str, str]]:
        items = {}
        for name, item in self.properties.items():
            items.update({name: item.Dict})
        return items
    
    def property_list(self) ->list[dict]:
        property_dict = self.get_items()
        property_list = []
        for name, item in property_dict.items():
            property_list.append({"name":name, "item":item})
        return property_list
    
#--------------------------[Database]---------------------#
# FC: [Database] object
@dataclass
class Database(RequestInfo): 
    object: str = 'database'
    description: list[RichText] = field(default_factory=list)
    is_inline: bool = False
    properties: dict[str, PropertyType | None] = field(default_factory=dict)
    title: list[RichText] = field(default_factory=list)
    
    def __post_init__(self) -> None:
        new_description = []
        for _, text in enumerate(self.description):
            if type(text) == dict:
                new_description.append(RichText(**text)) # type: ignore
        self.description = new_description
        
        if type(self.properties) == dict:
            new_properties = {}
            for name, value in self.properties.items():
                new_properties.update({name:PropertyType(**value)}) # type: ignore
            self.properties = new_properties
            
        new_title = []
        for _, text in enumerate(self.title):
            if type(text) == dict:
                new_title.append(RichText(**text)) # type: ignore
        self.title = new_title
        super().__post_init__()
        
    def asdict(self) -> dict:
        Dict = super().asdict()
        
        new_properties = {}
        for name, value in self.properties.items():
            if type(value) == dict or value is None:
                new_properties.update({name:value}) 
            else:
                new_properties.update({name:value.asdict()}) 
        Dict['properties'] = new_properties

        return Dict
    
    def update_title(self, title:RichText):
        self.title = [title]
        self.update()
        
    def update_property(self, name:str, **kwargs):
        property_type_object = self.properties[name]
        if property_type_object is not None:
            property_type = property_type_object.type
            self.properties[name] = PropertyType(name=name, type=property_type, **kwargs)
        else:
            self.properties[name] = PropertyType(name=name, **kwargs)
        self.update()
    
    def rename_property(self, old_name:str, new_name:str):
        self.properties[old_name].name = new_name # type: ignore
        self.update()

    def property_list(self) ->list[dict]:
        return [type_object.label_dict() if type_object is not None else None
                for _, type_object in self.properties.items()] # type: ignore
  
    def del_property(self, name:str):
        self.properties[name] = None
        self.update()
        
# FC: [Database] container object: paginated responses
@dataclass
class DatabaseContainer(BaseMethod):
    has_more: bool = False
    next_cursor: str | None = None
    object: str = 'list'
    page: dict = field(default_factory=dict)
    results: list[Page] = field(default_factory=list)
    type: str = field(default_factory=str)
    
    def __post_init__(self) -> None:
        new_pages = []
        for page in self.results:
            if type(page) == dict:
                new_pages.append(Page(**page)) # type: ignore
            else:
                new_pages.append(page)
        self.results = new_pages
        super().__post_init__()
        
    def asdict(self) -> dict:
        Dict = super().asdict()

        Dict['results'] = [page if type(page) == dict else page.asdict()
                            for page in self.results]
        return Dict

#--------------------------[Block]---------------------#
# FC: [Block] Retrieve request info
@dataclass
class Block(BlockBase, BaseMethod):
    type: str = field(default_factory=str)
    
    # text block
    paragraph: Paragraph = field(default_factory=Paragraph)
    heading_1: Heading = field(default_factory=Heading)
    heading_2: Heading = field(default_factory=Heading)
    heading_3: Heading = field(default_factory=Heading)
    callout: Callout = field(default_factory=Callout)
    quote: Quote = field(default_factory=Quote)
    bulleted_list_item: BulletedListItem = field(default_factory=BulletedListItem)
    numbered_list_item: NumberedListItem = field(default_factory=NumberedListItem)
    to_do: ToDo = field(default_factory=ToDo)
    toggle: Toggle = field(default_factory=Toggle)
    code: Code = field(default_factory=Code)
    # child page
    child_page: ChildPage = field(default_factory=ChildPage)
    child_database: ChildDatabase = field(default_factory=ChildDatabase)
    # Url block
    embed: Url = field(default_factory=Url)
    bookmark: Url = field(default_factory=Url)
    link_preview: Url = field(default_factory=Url)
    ## Media block
    image: FileLink = field(default_factory=FileLink)
    video: FileLink = field(default_factory=FileLink)
    pdf: FileLink = field(default_factory=FileLink)
    # Other
    equation: Expression = field(default_factory=Expression)
    divider: dict = field(default_factory=dict)
    table_of_contents: ColorMethod = field(default_factory=ColorMethod)
    breadcrumb: dict = field(default_factory=dict)
    template: Template = field(default_factory=Template)
    link_to_page: LinkTo = field(default_factory=LinkTo)
    synced_block: SyncedBlock = field(default_factory=SyncedBlock)
    ## column is column_list children.
    column_list: dict = field(default_factory=dict)
    column: dict = field(default_factory=dict)
    ## table_row is table children.
    table: Table = field(default_factory=Table)
    table_row: TableRow = field(default_factory=TableRow)
    
    def __post_init__(self) -> None:
        if not self.type in BLOCK_TYPE_LIST:
            raise ValueError(f"Block type {self.type} is not in the BLOCK_TYPE_LIST")
        
        self.paragraph = Paragraph(**self.paragraph) if type(self.paragraph) == dict else self.paragraph # type: ignore
        self.heading_1 = Heading(**self.heading_1) if type(self.heading_1) == dict else self.heading_1 # type: ignore
        self.heading_2 = Heading(**self.heading_2) if type(self.heading_2) == dict else self.heading_2 # type: ignore
        self.heading_3 = Heading(**self.heading_3) if type(self.heading_3) == dict else self.heading_3 # type: ignore
        self.callout = Callout(**self.callout) if type(self.callout) == dict else self.callout # type: ignore
        self.quote = Quote(**self.quote) if type(self.quote) == dict else self.quote # type: ignore
        self.bulleted_list_item = BulletedListItem(**self.bulleted_list_item) if type(self.bulleted_list_item) == dict else self.bulleted_list_item # type: ignore
        self.numbered_list_item = NumberedListItem(**self.numbered_list_item) if type(self.numbered_list_item) == dict else self.numbered_list_item # type: ignore
        self.to_do = ToDo(**self.to_do) if type(self.to_do) == dict else self.to_do # type: ignore
        self.toggle = Toggle(**self.toggle) if type(self.toggle) == dict else self.toggle # type: ignore
        self.code = Code(**self.code) if type(self.code) == dict else self.code # type: ignore
        self.child_page = ChildPage(**self.child_page) if type(self.child_page) == dict else self.child_page # type: ignore
        self.child_database = ChildDatabase(**self.child_database) if type(self.child_database) == dict else self.child_database # type: ignore
        self.embed = Url(**self.embed) if type(self.embed) == dict else self.embed # type: ignore
        self.bookmark = Url(**self.bookmark) if type(self.bookmark) == dict else self.bookmark # type: ignore
        self.link_preview = Url(**self.link_preview) if type(self.link_preview) == dict else self.link_preview # type: ignore
        self.image = FileLink(**self.image) if type(self.image) == dict else self.image # type: ignore
        self.video = FileLink(**self.video) if type(self.video) == dict else self.video # type: ignore
        self.pdf = FileLink(**self.pdf) if type(self.pdf) == dict else self.pdf # type: ignore
        self.equation = Expression(**self.equation) if type(self.equation) == dict else self.equation # type: ignore
        self.table_of_contents = ColorMethod(**self.table_of_contents) if type(self.table_of_contents) == dict else self.table_of_contents # type: ignore
        self.template = Template(**self.template) if type(self.template) == dict else self.template # type: ignore
        self.link_to_page = LinkTo(**self.link_to_page) if type(self.link_to_page) == dict else self.link_to_page # type: ignore
        self.synced_block = SyncedBlock(**self.synced_block) if type(self.synced_block) == dict else self.synced_block # type: ignore
        self.table = Table(**self.table) if type(self.table) == dict else self.table # type: ignore
        self.table_row = TableRow(**self.table_row) if type(self.table_row) == dict else self.table_row # type: ignore
        super().__post_init__()

    def __zip(self) -> dict:
        return dict(zip(BLOCK_TYPE_LIST,
                        [self.paragraph, self.heading_1, self.heading_2, self.heading_3,
                         self.callout, self.quote, self.bulleted_list_item, self.numbered_list_item,
                         self.to_do, self.toggle, self.code, self.child_page, self.child_database,
                         self.embed, self.bookmark, self.link_preview, self.image, self.video, self.pdf,
                         self.equation, self.divider, self.table_of_contents, self.breadcrumb, self.template, 
                         self.link_to_page, self.synced_block, self.column_list, self.column,
                         self.table, self.table_row]))
    
    def update(self) -> None:
        self.Dict = self.asdict()
   
    def asdict(self) -> dict:
        _zip_value = self.__zip()

        value = _zip_value[self.type].asdict() if _zip_value[self.type] != {} and type(_zip_value[self.type]) != dict else _zip_value[self.type]
        
        Dict = {
            "archived": self.archived,
            "created_by": self.created_by if type(self.created_by) == dict else self.created_by.asdict(),
            "created_time": self.created_time,
            "has_children": self.has_children,
            "id": self.id,
            "last_edited_by": self.last_edited_by if type(self.last_edited_by) == dict else self.last_edited_by.asdict(),
            "last_edited_time": self.last_edited_time,
            "object": self.object,
            "parent": self.parent if type(self.parent) == dict else self.parent.asdict(),
            "type": self.type,
            self.type: value,
        }

        return Dict

    def update_block(self, **kwargs):
        self.__init__(type=self.type,
                      id=self.id,
                      has_children=self.has_children,
                      archived=self.archived,
                      parent=self.parent,
                      **kwargs)
        self.__post_init__()
        
    def block_base_dict(self) -> dict:
        block_dict = self.Dict.copy()
        block_dict.pop('created_by')
        block_dict.pop('created_time')
        block_dict.pop('id')
        block_dict.pop('last_edited_by')
        block_dict.pop('last_edited_time')
        block_dict.pop('parent')
        return block_dict
    
    def block_item(self) -> dict:
        return {self.type: self.Dict[self.type]}
    
    def block_id_type(self) -> dict:
        return {
                "id":self.id,
                "type":self.type,
                }
        
# FC: [Block] children list
@dataclass
class BlockList(BaseMethod):
    """
    Raises:
        ValueError:  If the "type" is not in the BLOCK_CHILDREN_TYPES
    """
    block: dict = field(default_factory=dict)
    has_more: bool = False
    next_cursor: str | None = None
    object: str = 'list'
    results: list[Block] = field(default_factory=list)
    type: str = field(default_factory=str)
    
    def __post_init__(self) -> None:
        if not self.type in BLOCK_CHILDREN_TYPES:
            raise ValueError(f'Invalid type: {self.type}')
        
        new_results = []
        for block in self.results:
            if type(block) == dict:
                new_results.append(Block(**block)) # type: ignore
            else:
                new_results.append(block)
        self.results = new_results
        super().__post_init__()
    
    def asdict(self) -> dict:
        Dict = super().asdict()
        
        new_results = []
        for block in self.results:
            if type(block) == dict:
                new_results.append(block)
            else:
                new_results.append(block.asdict())
        Dict['results'] = new_results
        return Dict
    
    def get_block(self, id: int) -> Block | None:
        for block in self.results:
            if block.id == id:
                return block
        return None
    
    def get_block_short_list(self) -> list[dict]:
        return [block.block_id_type() for block in self.results]
        


####################################################################################
# MIT License
#
# Copyright (c) 2023 bluewhitep
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
####################################################################################
import logging

from notion_client import Client, AsyncClient
from notion_client.helpers import get_id

from .gadget import Gadget
from .api import Page, Database, User, Block
    
class kit:
    @staticmethod
    def get_id(url:str) ->str:
        """
        Get id from notion url

        Parameter
            url:            (str)       - Notion url
        """
        return get_id(url)
    
    @classmethod
    def method(cls) -> None:
        cls.Gadget:Gadget = Gadget()
        cls.Page:Page = Page(cls.client, id=None)         
        cls.Database:Database = Database(cls.client, id=None)
        cls.User:User = User(cls.client, id=None)
        cls.Block = Block(cls.client, id=None)
    
    @classmethod
    def Client(cls, token:str, log_level:int=logging.WARNING) ->Client:
        """
        Set client from notion_client

        Parameter
            token:          (str)       - Notion token
            log_level:      (int)       - Log level
        """
        cls.token = token
        cls.log_level = log_level
        cls.client = Client(auth=token, log_level=log_level)
        
        cls.method()
        return  cls.client
        
    @classmethod
    def Async_client(cls, token:str, log_level:int=logging.WARNING) ->AsyncClient:
        cls.token = token
        cls.log_level = log_level
        cls.client = AsyncClient(auth=token, log_level=log_level)
        
        cls.method()
        return  cls.client


####################################################################################
# MIT License
#
# Copyright (c) 2023 bluewhitep
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
####################################################################################
from pprint import pformat

from notion_kit import object
from notion_kit.dict_gadget import Dict_Gadget

class Debug:
    @staticmethod
    def fill_line(str_list:list, max_line:int, width:int=0) ->None:
        """
        Fill line with spaces

        Parameter
            str_list:       (list)      - list of string    
            max_line:       (int)       - max line
            width:          (int)       - width of each line. default 0
        """
        str_len = len(str_list)
        fill_count = max_line - str_len
        fill_line = "" if width == 0 else " "*width
        for _ in range(fill_count):
            str_list.append(fill_line)
    
    @staticmethod  
    def print_both(object1, object2, each_width=80, seq="#", grid:bool=False) ->None:
        """
        Print two object side by side
        
        Parameter
            object1:        (object)    - object 1
            object2:        (object)    - object 2
            each_width:     (int)       - pformat width. default 80
            seq:            (str)       - sequence between two object. default "#"
            grid:           (bool)      - show grid. default False
        """
        str1 = pformat(object1, width=each_width, sort_dicts=True).splitlines()
        str2 = pformat(object2, width=each_width, sort_dicts=True).splitlines()
        
        max_width = max([len(line) for line in str1]) + 1
        max_line = max(len(str1), len(str2))
        Debug.fill_line(str1, max_line, max_width)
        Debug.fill_line(str2, max_line)
        
        if grid:
            number="123456789|"*((max_width*2)//10 +1)
            number = number[:max_width*2+1]
            print(number + "| Total: " + str(len(number)))
            
        for i in range(max_line):
            filled_str1 = str1[i]+" "*(max_width - len(str1[i]))
            line = filled_str1 + seq + " " + str2[i]
            print(line)

    @staticmethod
    def print_diff(object1, object2, each_width=80, seq="#", grid:bool=False) ->None:
        """
        Print the difference
        
        Parameter
            object1:        (object)    - object 1
            object2:        (object)    - object 2
            each_width:     (int)       - pformat width. default 80
            seq:            (str)       - sequence between two object. default "#"
            grid:           (bool)      - show grid. default False
        """
        str1 = pformat(object1, width=each_width, sort_dicts=True).splitlines()
        str2 = pformat(object2, width=each_width, sort_dicts=True).splitlines()
        
        max_width = max([len(line) for line in str1]) + 1
        max_line = max(len(str1), len(str2))
        Debug.fill_line(str1, max_line, max_width)
        Debug.fill_line(str2, max_line)
        
        if grid:
            number="123456789|"*((max_width*2)//10 +1)
            number = number[:max_width*2+1]
            print(" |"+ number + "| Total: " + str(len(number)))
            
        for i in range(max_line):
            is_different = str1[i] != str2[i]
            BG_RED = '\033[41m'
            RESET = '\033[0m'
            marker = "-|"if is_different else " |" 
            filled_str1 = str1[i]+" "*(max_width - len(str1[i]))
            line = marker+ filled_str1 + seq + " " + str2[i]
            line = BG_RED + line + BG_RED if is_different else RESET + line + RESET
            print(line)

class Object:
    #-----------------------[Get object]-----------------------#
    # FC: [Date]
    @staticmethod
    def get_date(start:str, end:str | None = None) ->object.Date:
        """
        Get date
        
        Parameter
            start:              (str)       - start
            end:                (str)       - end
        
        Return
            object.Date
        """
        return object.Date(start=start, end=end)
    
    # FC: [File Link]
    @staticmethod
    def get_file_link(name:str, url:str) ->object.FileLink:
        """
        Get file link
        
        Parameter
            name:               (str)       - name
            url:                (object.Url) - url
        
        Return
            object.FileLink
        """
        return object.FileLink(name=name, external=object.Url(url=url))
    
    # FC: [File Upload]
    @staticmethod
    def get_file_upload(name:str, url:str, expiry_time:str) ->object.FileUpload:
        """
        Get file upload
        
        Parameter
            name:               (str)       - name
            url:                (object.Url) - url
        
        Return
            object.FileUpload
        """
        return object.FileUpload(name=name, 
                                 file=object.Upload(url=url, 
                                                    expiry_time=expiry_time))
        
    # FC: [Formula Value]
    @staticmethod
    def get_formula_value(type:str, 
                          string:str | None = None,
                          number:float | None = None,
                          boolean:bool = False,
                          date_start:str | None = None,
                          date_end:str | None = None) ->object.FormulaValue:
        """
        Get formula value
        
        Parameter
            type:               (str)       - type
            string:             (str)       - string
            number:             (float)     - number
            boolean:            (bool)      - boolean
            date_start:         (str)       - date start
            date_end:           (str)       - date end
            
        Return
            object.FormulaValue
        
        Raise
            ValueError      - date_start is required
            Invalid type    - type is invalid
        """
        if type == "string":
            return object.FormulaValue(type=type, string=string)
        elif type == "number":
            return object.FormulaValue(type=type, number=number)
        elif type == "boolean":
            return object.FormulaValue(type=type, boolean=boolean)
        elif type == "date":
            assert date_start is not None, "date_start is required"
            return object.FormulaValue(type=type, date=object.Date(start=date_start,
                                                                   end=date_end))
        else:
            raise ValueError("Invalid type")
        
    # FC: [Option]
    @staticmethod
    def get_option(name:str, color:str="default") ->object.Option:
        """
        Get option
        
        Parameter
            name:               (str)       - name
            color:              (str)       - color
        
        Return
            object.Option
        """
        return object.Option(name=name, color=color)
    
    # FC: [Rich Text]
    @staticmethod
    def get_rich_text(text:str = "",
                      link:str | None = None,
                      bold: bool = False,
                      italic: bool = False,
                      strikethrough: bool = False,
                      underline: bool = False,
                      code: bool = False,
                      color: str = "default") ->object.RichText:
        """
        Get rich text
        
        Parameter
            text:               (str)       - text
            link:               (str)       - link
            bold:               (bool)      - bold
            italic:             (bool)      - italic
            strikethrough:      (bool)      - strikethrough
            underline:          (bool)      - underline
            code:               (bool)      - code
            color:              (str)       - color
        
        Return
            object.RichText
        """
        url = object.Url(url=link) if link is not None else None
        text_content = object.TextContent(content=text,
                                          link=url)
        text_style = object.TextStyle(bold, italic, strikethrough,
                                      underline, code, color)
        
        return object.RichText(text=text_content, annotations=text_style)
    
    # FC: [Hyperlink Rich Text]
    @staticmethod
    def get_hyperlink(text:str,
                      link:str,
                      bold: bool = False,
                      italic: bool = False,
                      strikethrough: bool = False,
                      underline: bool = False,
                      code: bool = False,
                      color: str = "default") ->object.RichText:
        """
        Get rich text
        
        Parameter
            text:               (str)       - text
            link:               (str)       - link
            bold:               (bool)      - bold
            italic:             (bool)      - italic
            strikethrough:      (bool)      - strikethrough
            underline:          (bool)      - underline
            code:               (bool)      - code
            color:              (str)       - color
        
        Return
            object.RichText
        """
        url = object.Url(url=link)
        text_content = object.TextContent(content=text,
                                          link=url)
        text_style = object.TextStyle(bold, italic, strikethrough,
                                      underline, code, color)
        
        return object.RichText(text=text_content, annotations=text_style)
    
    # FC: [Rollup Item]
    @staticmethod
    def get_rollup_item(type:str,
                        function:str,
                        number:int = 0,
                        date_start:str | None = None,
                        date_end:str | None = None,
                        array: list[dict] = [],
                        incomplete: dict = {}) -> object.RollupItem:
        """
        Get rollup item
        
        Parameter
            type:               (str)       - type
            function:           (str)       - function
            number:             (int)       - number
            data_start:         (str)       - data start
            data_end:           (str)       - data end
            array:              (list)      - array
            incomplete:         (dict)      - incomplete
            
        Return
            object.RollupItem
            
        Raise
            ValueError      - type is invalid
            ValueError      - date_start is required
        """
        if type == "number":
            return object.RollupItem(type=type, function=function,
                                     number=number)
        elif type == "date":
            assert date_start is not None, "date_start is required"
            return object.RollupItem(type=type, function=function,
                                     date=object.Date(start=date_start,
                                                      end=date_end))
        elif type == "array":
            return object.RollupItem(type=type, function=function,
                                     array=array)
        elif type == "incomplete":
            return object.RollupItem(type=type, function=function,
                                     incomplete=incomplete)
        else:
            raise ValueError("Invalid type")

class PropertyType:
    #-----------------------[Property Type]-----------------------#
    # FC: [Generic Property Type]
    @staticmethod
    def get(name:str, type:str) ->object.PropertyType:
        """
        Get property type
        
        Parameter
            name:       (str)       - name
            type:       (str)       - type
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type=type)
    
    # FC: [Rich text]
    @staticmethod
    def get_rich_text(name:str) ->object.PropertyType:
        """
        Get rich text property type
        
        Parameter
            name:       (str)       - name
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type='rich_text')
    
    # FC: [Number]
    @staticmethod
    def get_number(name:str, format:str='number') ->object.PropertyType:
        """
        Get number property type
        
        Parameter
            name:       (str)       - name
            format:     (str)       - format. default "number"
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type='number', 
                                   number=object.Number(format=format))
    
    # FC: [Select]
    @staticmethod
    def get_select(name:str, options:list[object.Option]) ->object.PropertyType:
        """
        Get select property type
        
        Parameter
            name:       (str)       - name
            options:    (list[object.Option]) - options
            
        Return
            object.PropertyType
            
        Raise
            AssertionError - options must be list of object.Option
        """
        for option in options:
            assert type(option) == object.Option, "options must be list of object.Option"
        return object.PropertyType(name=name, type='select', 
                                   select=object.Options(options=options))
    
    # FC: [Multi-select]
    @staticmethod
    def get_multi_select(name:str, options:list[object.Option]) ->object.PropertyType:
        """
        Get multi-select property type
        
        Parameter
            name:       (str)       - name
            options:    (list[object.Option]) - options
            
        Return
            object.PropertyType
            
        Raise
            AssertionError - options must be list of object.Option
        """
        for option in options:
            assert type(option) == object.Option, "options must be list of object.Option"
        return object.PropertyType(name=name, type='multi_select', 
                                   multi_select=object.Options(options=options))
        
    # FC: [Date]
    @staticmethod
    def get_date(name:str) ->object.PropertyType:
        """
        Get date property type
        
        Parameter
            name:       (str)       - name
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type='date')
    
    # FC: [People]
    @staticmethod
    def get_people(name:str) ->object.PropertyType:
        """
        Get people property type
        
        Parameter
            name:       (str)       - name
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type='people')
    
    # FC: [Files]
    @staticmethod
    def get_files(name:str) ->object.PropertyType:
        """
        Get files property type
        
        Parameter
            name:       (str)       - name
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type='files')
    
    # FC: [Checkbox]
    @staticmethod
    def get_checkbox(name:str) ->object.PropertyType:
        """
        Get checkbox property type
        
        Parameter
            name:       (str)       - name
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type='checkbox')
    
    # FC: [Url]
    @staticmethod
    def get_url(name:str) ->object.PropertyType:
        """
        Get url property type
        
        Parameter
            name:       (str)       - name
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type='url')
    
    # FC: [Email]
    @staticmethod
    def get_email(name:str) ->object.PropertyType:
        """
        Get email property type
        
        Parameter
            name:       (str)       - name
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type='email')
    
    # FC: [Phone Number]
    @staticmethod
    def get_phone_number(name:str) ->object.PropertyType:
        """
        Get phone number property type
        
        Parameter
            name:       (str)       - name
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type='phone_number')
    
    # FC: [Relation Single]
    @staticmethod
    def get_relation_single(name:str, database_id:str) ->object.PropertyType:
        """
        Get relation single property type
        
        Parameter
            name:           (str)       - name
            database_id:    (str)       - database id
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type='relation', 
                                   relation=object.RelationSingle(database_id=database_id))
    
    # FC: [Relation Dual]
    @staticmethod
    def get_relation_dual(name:str, database_id:str,
                                    synced_id:str,
                                    synced_name:str) ->object.PropertyType:
        """
        Get relation dual property type
        
        Parameter
            name:           (str)       - name
            database_id:    (str)       - database id
            
        Return
            object.PropertyType
        """
        dual_property = object.RelationSynced(synced_property_id=synced_id,
                                             synced_property_name=synced_name)
        relation_dual = object.RelationDual(database_id=database_id,
                                            dual_property=dual_property)
        return object.PropertyType(name=name, type='relation', 
                                   relation=relation_dual)
    
    # FC: [Formula]
    @staticmethod
    def get_formula(name:str, expression:str="") ->object.PropertyType:
        """
        Get formula property type
        
        Parameter
            name:           (str)       - name
            expression:     (str)       - expression
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type='formula', 
                                   formula=object.Formula(expression=expression))
    
    # FC: [Created Time]
    @staticmethod
    def get_created_time(name:str) ->object.PropertyType:
        """
        Get created time property type
        
        Parameter
            name:       (str)       - name
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type='created_time')
    
    # FC: [Created By]
    @staticmethod
    def get_created_by(name:str) ->object.PropertyType:
        """
        Get created by property type
        
        Parameter
            name:       (str)       - name
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type='created_by')
    
    # FC: [Last Edited Time]
    @staticmethod
    def get_last_edited_time(name:str) ->object.PropertyType:
        """
        Get last edited time property type
        
        Parameter
            name:       (str)       - name
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type='last_edited_time')
    
    # FC: [Last Edited By]
    @staticmethod
    def get_last_edited_by(name:str) ->object.PropertyType:
        """
        Get last edited by property type
        
        Parameter
            name:       (str)       - name
            
        Return
            object.PropertyType
        """
        return object.PropertyType(name=name, type='last_edited_by')
    
    # FC: [Delete]
    @staticmethod
    def delete(name:str) ->dict:
        """
        Get delete property type
        
        Parameter
            name:       (str)       - name
            
        Return
            dict
        """
        return {name: None}
    
    # FC: [Rename]
    @staticmethod
    def rename(old_name:str, new_name:str) ->dict:
        """
        Get rename property type
        
        Parameter
            old_name:   (str)       - old name
            new_name:   (str)       - new name
            
        Return
            dict
        """
        return {old_name: {"name":new_name}}
    
class PropertyItem:
    #-----------------------[Property Item]-----------------------#
    pass

class Block:
    #-----------------------[Block]-----------------------#
    # FUNTURE: [Get block tree]
    pass

class Gadget:
    Debug = Debug()
    Object = Object()
    PropertyType = PropertyType()
    # PropertyItem = PropertyItem()
    # Block = Block()
    Dict = Dict_Gadget()
    
    #-----------------------[Database]-----------------------#
    # FC: [Update database property]
    @staticmethod
    def update_database_property(database_object:object.Database,
                                 property_name:str,
                                 new_property:object.PropertyType) ->object.Database:
        """
        Update database property
        
        Parameter
            database_object:    (object.Database)       - database object
            property_name:      (str)                   - property name
            new_property:       (object.PropertyType)   - new property
            
        Return
            object.Database [option]
        """
        new_property.type = database_object.properties[property_name].type # type: ignore
        new_property.id = database_object.properties[property_name].id # type: ignore
        new_property.name = database_object.properties[property_name].name # type: ignore
        database_object.properties[property_name] = new_property
        database_object.update()
        return database_object
    
    # FC: [Update database title]
    @staticmethod
    def update_database_title(database_object:object.Database,
                              **kwargs) ->object.Database:
        """
        Update database title
        
        Parameter
            database_object:    (object.Database)       - database object
            kwargs:             (dict)                  - kwargs
            
        Return
            object.Database [option]
        """
        rich_text = Object.get_rich_text(**kwargs)
        database_object.update_title(rich_text)
        return database_object
    
    # FC: [Update database icon]
    @staticmethod
    def update_database_icon(database_object:object.Database,
                            emoji:str) ->object.Database:  
        """
        Update database icon
        
        Parameter
            database_object:    (object.Database)       - database object
            icon:                (str)                   - icon
            
        Return
            object.Database [option]
        """
        database_object.icon = object.Icon(emoji=emoji)
        return database_object
        
    # FC: [Rename database property]
    @staticmethod
    def rename_database_property(database_object:object.Database,
                                 old_name:str,
                                 new_name:str) ->object.Database:
        """
        Rename database property
        
        Parameter
            database_object:    (object.Database)       - database object
            old_name:            (str)                   - old name
            new_name:            (str)                   - new name
            
        Return
            object.Database [option]
        """
        database_object.rename_property(old_name, new_name)
        return database_object
    
    # FC: [Delete database property]
    @staticmethod
    def del_database_property(database_object:object.Database,
                                property_name:str) ->object.Database:
        """
        Delete database property
        
        Parameter
            database_object:    (object.Database)       - database object
            property_name:      (str)                   - property name
            
        Return
            object.Database [option]
        """
        database_object.del_property(property_name)
        return database_object
    
    #-----------------------[Page]-----------------------#
    # FC: [Update page property item]
    @staticmethod
    def update_page_property(page_object:object.Page,
                             property_name:str,
                             **kwargs) ->object.Page:
        """
        Update page property
        
        Parameter
            page_object:    (object.Page)           - page object
            property_name:      (str)                   - property name
            new_property:       (object.PropertyItem)   - new property
            
        Return
            object.Page [option]
        """
        new_property = object.PropertyItem(id=page_object.properties[property_name].id,
                                           type=page_object.properties[property_name].type,
                                           **kwargs)
        page_object.properties[property_name] = new_property
        page_object.update()
        return page_object
    
    # FC: [Update page title]
    @staticmethod
    def update_page_title(page_object:object.Page,
                          **kwargs) ->object.Page:
        """
        Update page title
        
        Parameter
            page_object:    (object.Page)           - page object
            kwargs:             (dict)                  - kwargs
            
        Return
            object.Page [option]
        """
        rich_text = Object.get_rich_text(**kwargs)
        page_object.update_title(rich_text)
        return page_object
    
    # FC: [Upadate page icon]
    @staticmethod
    def update_page_icon(page_object:object.Page,
                         emoji:str) ->object.Page:
        """
        Update page icon
        
        Parameter
            page_object:    (object.Page)           - page object
            icon:                (str)                   - icon
            
        Return
            object.Page [option]
        """
        page_object.icon = object.Icon(emoji=emoji)
        return page_object
    
    # FC: [Clear page property]
    @staticmethod
    def clear_page_property(page_object:object.Page,
                            property_name:str) ->object.Page:
        """
        Clear page property
        
        Parameter
            page_object:    (object.Page)           - page object
            property_name:      (str)                   - property name
            
        Return
            object.Page [option]
        """
        page_object.clear_item(property_name)
        return page_object



####################################################################################
# MIT License
#
# Copyright (c) 2023 bluewhitep
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
####################################################################################

TEXT_COLOR_LIST=[
                 "blue",    "blue_background",
                 "brown",   "brown_background",
                 "default",
                 "gray",    "gray_background",
                 "green",   "green_background",
                 "orange",  "orange_background",
                 "pink",    "pink_background",
                 "purple",  "purple_background",
                 "red",     "red_background",
                 "yellow",  "yellow_background"
]
     
#---------------------[Property]---------------------#            
ALL_PROPERTIES_TYPES = [
                        "title", "rich_text", "number",  "select", "multi_select",
                        "status", "date", "people", "files", "checkbox", "url", 
                        "email", "phone_number",  "relation", "formula", "rollup",
                        "created_time",  "created_by", 
                        "last_edited_time", "last_edited_by",
]

NON_CREATEABLE_PROPERTIES_TYPES = [
                                    "title",  # Because it is created by default
                                    "status", # Not Recommended [Non-update property type]. Because it is can not add Groups and Options 
                                    "rollup", # Because it is need know "relation_property_id" and "rollup_property_id"
]

NON_UPDATABLE_PROPERTIES_TYPES = [
                                    "status",  # Because it is not a valid property schema by Notion API
]

NON_UPDATABLE_PROPERTIES_ITEMS = [
                                  "created_time", "created_by",
                                  "last_edited_time", "last_edited_by",
                                  "formula", "rollup",
]

FORMAT_TYPES = [
                "number", "number_with_commas", "percent", "dollar",
                "canadian_dollar", "euro", "pound", "yen", "ruble", "rupee", 
                "won", "yuan", "real", "lira", "rupiah", "franc", 
                "hong_kong_dollar", "new_zealand_dollar", "krona", 
                "norwegian_krone", "mexican_peso", "rand", "new_taiwan_dollar", 
                "danish_krone", "zloty", "baht", "forint", "koruna", "shekel", 
                "chilean_peso", "philippine_peso", "dirham", "colombian_peso", 
                "riyal", "ringgit", "leu", "argentine_peso", "uruguayan_peso", 
                "singapore_dollar",
]

ROLLUP_FUNCTION_LIST = [
                        "count_all", "count_values", "count_unique_values", 
                        "count_empty", "count_not_empty", 
                        "percent_empty", "percent_not_empty", 
                        "sum", "average", "median", "min", "max", "range", 
                        "show_original",
]

ROLLUP_TYPE_LIST = [
                    "number", "date", "array", "unsupported", "incomplete"
]

#---------------------[Block]---------------------#

BLOCK_CHILDREN_TYPES = [
                        "block", 
                        "page", 
                        "user", 
                        "database", 
                        "property_item", 
                        "page_or_database"
]

BLOCK_TYPE_LIST = [
                    "paragraph", "heading_1", "heading_2", "heading_3", 
                    "callout", "quote", "bulleted_list_item",
                    "numbered_list_item", "to_do", "toggle", "code",
                    "child_page", "child_database",
                    "embed", "bookmark", "link_preview", 
                    "image", "video", "pdf", "equation",
                    "divider",
                    "table_of_contents", "breadcrumb",
                    "template", "link_to_page", "synced_block",
                    "column_list", "column", "table", "table_row",
]

CODE_LANGUAGE_LIST =[
                    "abap", "arduino", "bash", "basic", "c", "clojure", 
                    "coffeescript",  "c++", "c#", "css", "dart", "diff",
                    "docker", "elixir", "elm", "erlang", "flow", "fortran",
                    "f#", "gherkin", "glsl", "go", "graphql", "groovy",
                    "haskell", "html", "java", "javascript", "json",
                    "julia", "kotlin", "latex", "less", "lisp", "livescript",
                    "lua", "makefile", "markdown", "markup", "matlab",
                    "mermaid", "nix", "objective-c", "ocaml", "pascal",
                    "perl", "php", "plain text", "powershell", "prolog",
                    "protobuf", "python", "r", "reason", "ruby", "rust",
                    "sass", "scala", "scheme", "scss", "shell", "sql",
                    "swift", "typescript", "vb.net", "verilog", "vhdl",
                    "visual basic", "webassembly", "xml", "yaml",
                    "java/c/c++/c#",
]

MEDIA_TYPE_LIST = [
                    "image", "video", "file", "pdf",
]


from setuptools import setup

def _get_description(filename):
    with open(filename, encoding='utf-8') as file:
        return file.read()

def _requires_from_file(filename):
    with open(filename, encoding="utf-8") as file:
        return file.read().splitlines()

_pkg_name = "notion_kit"
_version = "1.0.2" # Change 1.0.1 => 1.0.2
_packages=["notion_kit"]
_install_requires=_requires_from_file("requirements.txt"),
_author = "Jieqiang Zhang"
_author_email = "bluewhite2389@gmail.com"
_description = "For easy use notion-sdk-py"
_long_description = _get_description("README.md")
_long_description_content_type = "text/markdown"
_url = "https://github.com/bluewhitep/Notion_kit"
_classifiers = [
    "Programming Language :: Python :: 3.10",
    "Development Status :: 3 - Alpha",
    "License :: OSI Approved :: MIT License",
]
_python_requires=">=3.10, <4"

setup(
    name=_pkg_name,
    version=_version,
    packages=_packages,
    install_requires=_install_requires,
    author=_author,
    author_email=_author_email,
    description=_description,
    license='MIT',
    long_description=_long_description,
    long_description_content_type=_long_description_content_type,
    url=_url,
    classifiers=_classifiers,
    python_requires=_python_requires,
)
